# -*- coding: utf-8 -*-
"""Description: More convenient way for accessing specified path or file locaiton.
It's written a bit earlier as this project builds. I only know ``os.path`` to get
all path at that time, but as I learned more knowledge of python, I think it will
be better if they are replaced with pathlib.Path and it will be more readable. 

TODO rewrite/simplify this module, use pathlib instead of os.path for better readability.

Author: BriFuture

Modified: 2019/03/10 19:57
"""
import os, sys

from os.path import dirname, abspath, pardir, expanduser, join, exists

CURRENT_PATH = dirname(abspath(__file__))
TOP_PATH = abspath(join(CURRENT_PATH, os.pardir))  # ./../
USER_HOME = expanduser("~")

DATA_DIR = abspath(join(USER_HOME, '.sspy-mgr'))
LOG_DIR = abspath(join(DATA_DIR, 'log'))

#配置文件的绝对路径 absolute path
CONFIG_PATH = abspath(join(DATA_DIR, 'config.yaml'))

if not exists(DATA_DIR):
    os.mkdir(DATA_DIR)
if not exists(LOG_DIR):
    os.mkdir(LOG_DIR)

import logging


def createLoggerFile(name, level=logging.DEBUG) -> logging.FileHandler:
    """Create log file under specified directory
    """
    fmt = "%(asctime)s %(levelname)s:  %(message)s"
    formatter = logging.Formatter(fmt)
    if len(name) == 0:
        name = 'root'
    fp = abspath(join(LOG_DIR, '{}.log'.format(name)))
    fh = logging.FileHandler(fp)
    fh.setLevel(level)
    fh.setFormatter(formatter)
    return fh

log_to_console = False
def createLogger(name, level=logging.DEBUG, stream=False, logger_prefix=""):
    """create logger for different modules
    @param ``name``: logger name. New Handlers won't be attached to the logger
    if logger already has Other Handlers.
    @param ``level`` set the lever of FileHandler, logging.DEBUG by default
    @param ``stream`` flag to indicate whether to add a StreamHandler, False by default
    """
    fh = createLoggerFile(name, level=level)

    logger = logging.getLogger(name)
    if len(logger.handlers) > 0:
        return logger
    logger.addHandler(fh)
    logger.setLevel(level)

    global log_to_console
    if stream or log_to_console:
        fmt = "%(asctime)s %(levelname)s: {} %(message)s".format(logger_prefix)
        formatter = logging.Formatter(fmt)
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        logger.addHandler(sh)

    return logger



# -*- coding: utf-8 -*-

from sspymgr import DB, createLogger
logger = createLogger("plugin_annoucement", stream=False, logger_prefix="[Plugin announcement]")

from datetime import datetime

class Announcement(DB.Model):
    """公告数据模型
    """
    __tablename__ = 'announcement'
    id = DB.Column(DB.Integer, primary_key = True)
    title = DB.Column(DB.String(255), nullable=False)
    variant = DB.Column(DB.String(32)) # primary, success, warn
    content = DB.Column(DB.BLOB)
    createtime = DB.Column(DB.DateTime, default=datetime.now)
    top = DB.Column(DB.Integer, default=0) 

    def to_dict(self):
        content = self.content.decode('utf-8') if self.content is not None else ''
        di = {
            'id': self.id,
            'title': self.title,
            'variant': self.variant,
            'content': content,
            'createTime': self.createtime.timestamp(),
            'top': self.top,
        }
        return di

app = None
from flask import request, jsonify
def registerApis(api):
    @api.route_admin('/announcement/create', methods=['POST'])
    def admin_createAnnouncement():
        title = request.form.get('title', None)
        state = {'status': 'success'}
        if title is None:
            state['status'] = 'fail'
            return jsonify(state)
        variant = request.form.get('variant')
        content = request.form.get('content')
        top = request.form.get('top', 0)
        ann = Announcement(title=title, 
            variant=variant, content=content.encode(), top=top)
        app.m_db.session.add(ann)
        app.m_db.session.commit()
        return jsonify(state)

    @api.route_admin('/announcement/getAll', methods=['GET', 'POST'])
    def admin_getAllAnnouncement():
        anns = Announcement.query.all()
        state = {'status': 'success'}
        state['announcements'] = [a.to_dict() for a in anns]
        return jsonify(state)
    
    @api.route_admin('/announcement/delete', methods=['POST'])
    def admin_delAnnouncement():
        id = request.form.get("id", -1)
        state = {'status': 'success'}
        try:
            id = int(id)
            ann = Announcement.query.filter_by(id=id).delete()
        except Exception as exc:
            state['status'] = 'fail'
            state['desc'] = 'Wrong Id or no such item'
            logger.warn('Invalid operation, {}'.format(exc))
            
        return jsonify(state)

    @api.route_user('/announcement/getAll', methods=['GET', 'POST'])
    def user_getAllAnnouncement():
        anns = Announcement.query.all()
        state = {'status': 'success'}
        state['announcements'] = [a.to_dict() for a in anns]
        return jsonify(state)

    logger.debug("api registered")

def init(iapp):
    global app
    app = iapp
    app.m_events.on("beforeRegisterApi", registerApis)
    logger.info("inited")


# -*- coding: utf-8 -*-

from sspymgr.models import DB
from datetime import datetime, timedelta


class Account(DB.Model):
    """
    Represents shadowsocks server account, port and password can not be None
    """
    __tablename__ = 'account'
    id = DB.Column(DB.Integer, primary_key=True, autoincrement=True)
    # corresponding with Table User, but can be created by manager
    userId = DB.Column(DB.Integer, unique=True)  
    data = DB.Column(DB.String(255))
    port = DB.Column(DB.Integer, nullable=False)
    password = DB.Column(DB.String(255), nullable=False)
    totalFlow = DB.Column(DB.BigInteger)
    expire = DB.Column(DB.DateTime)
    server = DB.Column(DB.String(255))
    method = DB.Column(DB.String(64), default='aes-256-cfb')
    # server = DB.Column( DB.String( 255 ) )  # no need for multi server now
    """
    @field type
        0 registered by email and have a 
    """
    type = DB.Column(DB.Integer)
    status = DB.Column(DB.String(32))
    createTime = DB.Column(DB.DateTime, default=datetime.now())

    def __init__(self, **kwargs):
        super(Account, self).__init__(**kwargs)

    def get_status(self):
        if self.status == 0:
            return 0
        return 0

    def is_flow_remaining(self):
        """If this account has remaining flow, True returned, otherwise False.
        """
        if not hasattr(self, 'account_flow'):
            self.account_flow = AccountFlow.query.filter_by(port=self.port).first()
        return self.totalFlow > self.account_flow.flow

    def clear_flow(self):
        """Clear flow usage stats
        """
        if not hasattr(self, 'account_flow'):
            self.account_flow = AccountFlow.query.filter_by(
                port=self.port).first()
        self.account_flow.flow = 0

    def get_used_flow(self):
        if hasattr(self, '_flow'):
            return self._flow.flow
        return None

    def __repr__(self):
        return '<Account {} status: {}>'.format(self.port, self.status)

    def to_dict(self):
        di = {
            'id': self.id,
            'userId': self.userId,
            'data': self.data,
            'port': self.port,
            'password': self.password,
            'totalFlow': self.totalFlow,
            'usedFlow': self.get_used_flow(),
            'server': self.server,
            'method': self.method,
            'type': self.type,
            'status': self.status,
            'expire': self.expire.timestamp(),
            'createTime': self.createTime.timestamp(),
        }
        return di


class AccountFlow(DB.Model):
    __tablename__ = 'accountFlow'

    port = DB.Column(DB.Integer, primary_key=True)
    flow = DB.Column(DB.BigInteger, default=0)  # flow used
    status = DB.Column(DB.String(255), default='checked')
    accountId = DB.Column(DB.Integer, unique=True)
    updateTime = DB.Column(
        DB.DateTime, default=datetime.now)  # means the account last connected
    autoBanTime = DB.Column(DB.DateTime)

    # serverId  = DB.Column( DB.Integer )
    checkTime = DB.Column(DB.DateTime)
    nextCheckTime = DB.Column(DB.DateTime)

    def consume_flow(self, flow: int):
        self.flow += flow
        self.updateTime = datetime.now()

    def __repr__(self):
        return '<Flow port: %d, flow: %d>' % (self.port, self.flow)

    def to_dict(self):
        di = {
            'port': self.port,
            'flow': self.flow,
            'status': self.status,
            'accountId': self.accountId,
            'updateTime': self.updateTime.timestamp(),
            'autoBanTime': self.autoBanTime.timestamp(),
            'checkTime': self.checkTime.timestamp(),
            # 'nextCheckTime': self.nextCheckTime.timestamp(),
        }
        return di


class Command(DB.Model):
    __tablename__ = 'command'
    id = DB.Column(DB.Integer, primary_key=True)
    code = DB.Column(DB.String(255))
    time = DB.Column(DB.String(255))




# -*- coding: utf-8 -*-
from sspymgr.models import DB
from datetime import datetime, timedelta

# class SaveFlow( DB.DB.Model ):
#     __tablename__ = 'saveFlow'
#     id = DB.DB.Column( DB.DB.Integer, primary_key = True, autoincrement = True )
#     port = DB.DB.Column( DB.DB.Integer )
#     flow = DB.DB.Column( DB.DB.BigInteger )
#     time = DB.DB.Column( DB.DB.DateTime )
#     accountId = DB.DB.Column( DB.DB.Integer )

# def genFlowRecord( tablename ):
#     def __repr__( self ):
#         return '<FlowRecord: %s>' % tablename

#     props = {
#         '__tablename__' : tablename,
#         'id' : DB.Column( DB.Integer, primary_key = True, autoincrement = True ),
#         'flow' : DB.Column( DB.BigInteger ),
#         'port' : DB.Column( DB.Integer ),
#         'time' : DB.Column( DB.DateTime ),
#         'accountId' : DB.Column( DB.Integer ),
#     }
#     newclass = type( tablename, ( DB.Model, ), props )
#     return newclass

# FlowRecord = genFlowRecord( 'flowRecord' )
# FlowRecord5Min = genFlowRecord( 'flowRecord5min' )
# FlowRecordHour = genFlowRecord( 'flowRecordHour' )
# FlowRecordDay  = genFlowRecord( 'flowRecordDay' )


class FlowRecord5Min(DB.Model):
    __tablename__ = 'flowRecord_5min'
    id = DB.Column(DB.Integer, primary_key=True)
    port = DB.Column(DB.Integer)
    # unit of bytes
    flow = DB.Column(DB.Integer)  
    # store date in the format: 2018-10-10
    date = DB.Column(DB.String, nullable=False) 
    # 0 - 23
    hour = DB.Column(DB.Integer, nullable=False)  
    # noraml minute here, gap is 5, from 0 to 55: 0 ==> [ 00:00 ~ 04:59 ]
    minute = DB.Column(DB.Integer, nullable=False) 

    def to_dict(self):
        di = {
            'flow': self.flow,
            'hour': self.hour,
            'minute': self.minute
        }
        return di

    def __repr__(self):
        return '<FlowRecord5Min %d: %d at %s %d>' % (self.port, self.flow,
                                                     self.date, self.hour)

from datetime import datetime, timedelta
from sspymgr import formatTime
from sspymgr.sscontroller import Account, AccountFlow
from sspymgr.core import User, UserType

import logging
logger = logging.getLogger("plugin_sscontroller")


class FlowRecordDay(DB.Model):
    __tablename__ = 'flowRecord_Day'
    id = DB.Column(DB.Integer, primary_key=True)
    port = DB.Column(DB.Integer)
    flow = DB.Column(DB.Integer)  # unit of bytes
    date = DB.Column(
        DB.String,
        nullable=False)  # store date in the format (year and month): 2018-10
    day = DB.Column(DB.Integer, nullable=False)  # store day, from 1 to 31

    def to_dict(self):
        di = {
            'flow': self.flow,
            'day': self.day,
        }
        return di

    def __repr__(self):
        return '<FlowRecordDay %d: %d at %s %d>' % (self.port, self.flow,
            self.date, self.day)


from . import logger
class FlowStats(object):
    """统计流量，将流量记录到数据库中
    """
    MINUTE = 5

    def __init__(self, db, *args, **kwargs):
        self.db = db
        self.min = {}
        # self.hour = {}
        self.day = {}
        return super().__init__(*args, **kwargs)

    def recordFlow(self, port_flow: dict):
        """将流量数据记录到 dict 中
        port_flow: { 
            45001: flow, 
            45002: Account
        }
        """
        for port in port_flow:
            flow = port_flow[port]
            if port not in self.min:
                self.min[port] = flow
            else:
                self.min[port] = self.min[port] + flow

        for port in self.min:
            if port not in self.day:
                self.day[port] = self.min[port]
            else:
                self.day[port] += self.min[port]

        logger.debug("5 min: {}".format(self.min))

    def record5minToDb(self):
        now = datetime.now()
        cond = (now.minute + 1) % self.MINUTE
        if( cond != 0 ):
            return
        gen = 0
        for port in self.min:
            if self.min[port] == 0:
                # no need to record
                continue
            self.genMinRecord(port, self.min[port], now)
            self.min[port] = 0
            gen += 1
        logger.debug("5 min record to db count: {}".format(gen))
        self.db.session.commit()

    def record1hourToDb(self):
        pass

    def record1dayToDb(self):
        gen = 0
        for port in self.day:
            if self.day[port] == 0:
                continue
            self.genDayRecord(port, self.day[port], datetime.now())
            self.day[port] = 0
            gen += 1
        logger.debug("day record to db count: {}".format(gen))
        self.db.session.commit()

    def genMinRecord(self, port: int, flow: int, time: datetime):
        """time is record time, but should be 
        """
        d = '{:%Y-%m-%d}'.format(time)
        record = FlowRecord5Min(
            port=port, flow=flow, date=d, hour=time.hour, minute=time.minute)
        self.db.session.add(record)
        return record

    def genDayRecord(self, port: int, flow: int, time: datetime):
        """time: record time
        """
        date = '{:%Y-%m}'.format(time)
        record = FlowRecordDay(port=port, flow=flow, date=date, day=time.day)
        self.db.session.add(record)
        return record
        

class AccountsChecker(object):
    app = None
    def __init__(self, app, *args, **kwargs):
        self.app = app

    def checkAllAccounts(self):
        """Action: check accounts status and userType
        """
        dbaccounts = Account.query.all()
        # accountFlow = AccountFlow.query.all()
        for acc in dbaccounts:
            # for flow in accountFlow:
            #     if acc.port == flow.port:
            #         acc._flow = flow
            #         break
            uid = acc.userId
            user = User.query.filter_by(id=uid).first()
            if user is None:
                continue
            now = datetime.now()
            if now < acc.expire: 
                if self.canSendLastDay(user) and (acc.expire - now) < timedelta(days=1):
                    self.sendLastDay(acc, user)
            else:
                user.type = UserType.INVALID.value
            
        self.app.m_db.session.commit()
        
    def sendLastDay(self, acc: Account, user: User):
        logger.debug("<{} : {}> will expire soon".format(user.email, user.comment))

        # content = """Hi, your account of SSPYMGR will be expired within one day (at {}).<br />
        # Please go to {} to rechart Your account.(Your ss account Port: {})
        # """.format(formatTime(acc.expire), acc.server, acc.port)
        content = """你好, 你的 SSPYMGR 账户将会在一天后过期 (即 {}).<br />
        请访问 {} 以重新激活你的账户.(你的 shadowsocks 账号端口是: {})
        """.format(formatTime(acc.expire), acc.server, acc.port)
        user.type = UserType.NOTIFIED.value
        subject = "账户过期提醒"
        self.app.m_emailManager.add_email(
            to=user.email,
            subject=subject,
            content=content,
            type="last day notify")


    def canSendLastDay(self, user):
        if user.type is None or len(user.type) == 0:
            return False
        utype = UserType(user.type)
        if (utype == UserType.BANNED or utype == UserType.TRIAL):
            return False
        if (utype != UserType.NOTIFIED):
            return True
        return False


# -*- coding: utf-8 -*-
"""Description: This module stores the settings of webserver provided by sspymgr into database.
The original purpose of this module is to store as much information as possible into a single table, 
But it is proved that it's much better to store settings with simple and enumerable types, such as
String and Number (most common). It will take some time to review this module and make some improvements.

Author: BriFuture

Date: 2019/03/18 21:42
"""
from sspymgr import DB, createLogger
from sqlalchemy.sql import text
from base64 import b64decode

logger = createLogger('core_settings', stream=False, logger_prefix="[Plugin Settings]")

db = None
class WebguiSetting( DB.Model ):
    """ TODO reconstruct this model
    Type (can be ignored, just used as utility):
        String: Value type by default, should be processed by other logic
        Image: Value are encoded with base64 format
        Number: Treat value as number( use float builtin function), 
        Boolean: are stored in number with only two candicates (0 or 1)
        Json: Value are encoded with json format
        Hidden: Not visible for even manager
    """
    __tablename__ = 'webguiSetting'
    id = DB.Column( DB.Integer, primary_key = True, autoincrement = True )
    key = DB.Column( DB.String( 255 ), unique = True )
    value = DB.Column( DB.Text )
    type = DB.Column( DB.String( 32 ), default = 'String' )

    def getTypedValue( self ):
        if not self.type:
            return self.value
        if self.type == 'Boolean':
            return self.value == '1'
        if self.type == 'Number':
            try:
                return int( self.value )
            except Exception:
                return float( self.value )
        return self.value
            

    @staticmethod
    def getSetting(key, default_value=None, default_type="String"):
        setting = WebguiSetting.query.filter_by(key=key).first()
        if setting is None and default_value is not None :
            setting = WebguiSetting(key=key, value=default_value, type=default_type)
            db.session.add(setting)
            db.session.commit()
        return setting

    def to_dict(self):
        di = {
            'id': self.id,
            'key': self.key,
            'value': self.value if type(self.value) is str else self.value.decode() ,
            'type': self.type,
        }
        return di

def __init_webgui_setting(db):
    """
    alipay_enabled: bool default false
    wechatpay_enabled: bool default false
    alipay_qrcode: image data with base64 encode
    wechatpay_qrcode: image data with base64 encode
    signup_enabled: bool default true
    signup_without_invitation: int  default 8
    signup_email_limit: int  default 2
    shadowsock_port_range_start: Number default '50000'
    """
    try:
        sql = 'SELECT count(*) FROM %s' % WebguiSetting.__tablename__
        res = db.engine.execute( text( sql ) ).first()
        if res[ 0 ] != 0:
            return
        settings = [
            ( "alipay_enabled",    "0", "Boolean" ),
            ( "wechatpay_enabled", "0", "Boolean" ),
            ( "signup_enabled", "1", "Boolean" ),
            ( "signup_without_invitation", "8", "Number" ),
            ( "signup_email_limit", "2", "Number" ),
            ( "shadowsock_port_range_start", "45000", "Number" ),
        ]
        for setting in settings:
            wgs = WebguiSetting( key = setting[0], value = setting[1], type = setting[2] )
            db.session.add( wgs )
        
        db.session.commit()
    except Exception:
        pass

from flask import jsonify

def registerApi(api, app):
    global db
    db = app.m_db
    __init_webgui_setting(db)
    # TODO change setting item
    @api.route_admin('/setting/getAll', methods=['POST'])
    def get_all_settings():
        all = WebguiSetting.query.all()
        settings = [s.to_dict() for s in all]
        # for s in all:
        #     settings.append(s.to_dict())
        return jsonify({
            'status': 'success',
            'setting': settings
        })
    logger.info("api registered.")


# -*- coding: utf-8 -*-
from sspymgr import getRandomCode, convertFlowToByte
from sspymgr.sscontroller import *
import logging
logger = logging.getLogger("plugin_sscontroller")


from datetime import datetime, timedelta
from flask import request, jsonify, session

from sqlalchemy.sql import func

from .flow_stats import FlowRecord5Min, FlowRecordDay

def registerRoutes(api, app):
    """register more rotues for sscontroller
    """
    @api.route_admin('/flowUsage/getInDay', methods=['GET', 'POST'])
    def admin_getFlowUsage_day():
        state = {'status': 'success'}
        accountId = request.form.get('accountId', -1)
        account = Account.query.filter_by(id = accountId).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            d = int(date[2])
            now = datetime(year=y, month=m, day=d)
        except:
            now = datetime.now()
        date = '{:%Y-%m-%d}'.format(now)
        records = FlowRecord5Min.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)
        
    @api.route_admin('/flowUsage/getInMon', methods=['GET', 'POST'])
    def admin_getFlowUsage_month():
        state = {'status': 'success'}
        accountId = request.form.get('accountId', -1)
        account = Account.query.filter_by(id = accountId).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            now = datetime(year=y, month=m)
        except:
            now = datetime.now()
        date = '{:%Y-%m}'.format(now)
        records = FlowRecordDay.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)

    @api.route_admin('/account/<id>/flowRecord_5min', methods=['GET', 'POST'])
    def get_account_flowRecord_5min(id):
        """ flow unit is KB
        """
        acc = Account.query.filter_by(id=id).first()
        if not acc:
            return '{}'

        now = datetime.now()
        date = '{:%Y-%m-%d}'.format(now)
        records = FlowRecord5Min.query.filter_by(port=acc.port, date=date).all()
        result = {}
        for r in records:
            key = r.hour * 60 + r.minute
            result[key] = r.flow / 1000
        return jsonify(result)

    @api.route_user('/flowUsage/getInDay', methods=['GET', 'POST'])
    def user_getFlowUsage_day():
        state = {'status': 'success'}
        uid = session.get('userid')
        account = Account.query.filter_by(userId=uid).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            d = int(date[2])
            now = datetime(year=y, month=m, day=d)
        except:
            now = datetime.now()
        date = '{:%Y-%m-%d}'.format(now)
        records = FlowRecord5Min.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)
        
    @api.route_user('/flowUsage/getInMon', methods=['GET', 'POST'])
    def user_getFlowUsage_month():
        state = {'status': 'success'}
        uid = session.get('userid')
        account = Account.query.filter_by(userId=uid).first()
        if account is None:
            state['status'] = 'fail'
            return jsonify(state)
        port = account.port
        date = request.form.get('date', None)
        try:
            date = date.split('-')
            y = int(date[0])
            m = int(date[1])
            now = datetime(year=y, month=m)
        except:
            now = datetime.now()
        date = '{:%Y-%m}'.format(now)
        records = FlowRecordDay.query.filter_by(port=port, date=date)
        state['port'] = port
        state['date'] = now.timestamp()
        state['records'] = [fr.to_dict() for fr in records]
        return jsonify(state)

    logger.info("routes api registered")




# -*- coding: utf-8 -*-
from sspymgr import DB, \
    createLogger, getRandomCode, formatTime

logger = createLogger("plugin_order", stream=False, logger_prefix="[Plugin Order]")

from datetime import datetime, timedelta
from sqlalchemy.sql import text


class Product(DB.Model):
    __tablename__ = 'product'
    id = DB.Column(DB.Integer, primary_key=True, autoincrement=True)
    # name = DB.Column( DB.String( 255 ) )
    type = DB.Column(DB.Integer)  # 0 or none stands by enabled
    flow = DB.Column(DB.Integer)  # unit: Mn
    price = DB.Column(DB.Float)  # unit: ￥
    enable = DB.Column(DB.Boolean, default=True)
    duration = DB.Column(DB.Integer)  # unit: seconds
    buffer_period = DB.Column(DB.Integer)  # unit: seconds

    @staticmethod
    def add_type(price: float, flow: int, duration: timedelta,
                 buffer_period: timedelta):
        dseconds = duration.total_seconds()
        bseconds = buffer_period.total_seconds()
        ot = Product(
            price=price, flow=flow, duration=dseconds, buffer_period=bseconds)
        return ot

    def to_dict(self):
        return {
            "id": self.id,
            "flow": self.flow,
            "price": self.price,
            "enable": self.enable,
            "duration": self.duration,
            "bufferPeriod": self.buffer_period
        }

    def to_dict(self):
        di = {
            "id": self.id,
            "flow": self.flow,
            "price": self.price,
            "enable": self.enable,
            "duration": self.duration,
            "bufferPeriod": self.buffer_period
        }
        return di

    def __repr__(self):
        return '<Product %d %s>' % (self.id, self.enable)

class ProductQrcode(DB.Model):
    __tablename__ = 'productQrcode'
    id = DB.Column(DB.Integer, primary_key=True)
    pid = DB.Column(DB.Integer, nullable=False)  # product Id
    category = DB.Column(DB.String(32), nullable=True) # such as alipay or wechatpay
    qrcode = DB.Column(DB.TEXT)  # base64 encoded image
    desc = DB.Column(DB.String(255))  
    path = DB.Column(DB.String(255)) # if save image as files, path to find it
    remark = DB.Column(DB.String(32)) # reserved

    def to_dict(self):
        di = {
            "id": self.id,
            "pid": self.pid,
            "category": self.category,
            "qrcode": self.qrcode,
            "desc": self.desc,
            "path": self.path,
            "remark": self.remark,      
        }
        return di


class Order(DB.Model):
    __tablename__ = 'order'
    id = DB.Column(DB.Integer, primary_key=True, autoincrement=True)
    uid = DB.Column(DB.Integer, nullable=False)
    pid = DB.Column(DB.Integer, nullable=False)
    server = DB.Column(DB.String(255))
    code = DB.Column(DB.String(32))
    alipay = DB.Column(DB.Float)
    wechatpay = DB.Column(DB.Float)
    orderTime = DB.Column(DB.DateTime)
    commentByUser = DB.Column(DB.Text, default='')
    # whether the Manager has confirmed this order
    checked = DB.Column(DB.Boolean, default=False)  
    # buffer period
    bufferPeriodExpire = DB.Column(DB.DateTime)  
    expire = DB.Column(DB.DateTime)

    @staticmethod
    def place_order(uid: int, product: Product, **kwargs):
        od = Order(uid=uid, pid=product.id)
        # od.code = getRandomCode( 32 )
        od.code = '{:%Y%m%d%H%M%S}{:1}'.format(datetime.now(),
                                               getRandomCode(18))
        od.commentByUser = kwargs.get('comment', '')
        now = datetime.now()
        od.orderTime = now
        od.checked = kwargs.get('checked', False)
        od.bufferPeriodExpire = timedelta(seconds=product.buffer_period) + now
        od.expire = timedelta(seconds=product.duration) + now
        app.m_db.session.add(od)
        app.m_db.session.commit()
        # return od


    def is_valid(self):
        now = datetime.now()
        if not self.checked:
            return now < self.bufferPeriodExpire
        return now < self.expire

    def check_it(self, alipay=0, wechatpay=0):
        # if not alipay
        self.checked = True
        self.alipay = alipay
        self.wechatpay = wechatpay

    def relateProduct(self):
        return Product.query.filter_by(id=self.productId).first()

    def to_dict(self):
        di = {
            'id': self.id,
            'uid': self.uid,
            'pid': self.pid,
            'server': self.server,
            'code': self.code,
            'alipay': self.alipay,
            'wechatpay': self.wechatpay,
            'orderTime': self.orderTime.timestamp(),
            'commentByUser': self.commentByUser,
            'checked': self.checked,
            'bufferPeriodExpire': self.bufferPeriodExpire.timestamp(),
            'expire': self.expire.timestamp(),
        }
        return di


def init_product_type():
    sql = 'SELECT count(*) FROM %s' % Product.__tablename__
    res = app.m_db.engine.execute(text(sql)).first()
    
    count = res[0]
    if count == 0:
        ptype = (
            (1,  8 * 1024, timedelta(days=7), timedelta(hours=4)), 
            (2, 18 * 1024, timedelta(days=15), timedelta(hours=8)),
            (2.5, 24 * 1024, timedelta(days=15), timedelta(hours=10)), 
            (4, 40 * 1024, timedelta(days=30), timedelta(days=1)),
            (5, 55 * 1024, timedelta(days=30), timedelta(days=1))
        )
        for ot in ptype:
            new_product = Product.add_type(ot[0], ot[1], ot[2], ot[3])
            app.m_db.session.add(new_product)
        app.m_db.session.commit()


from flask import jsonify, session, request
from sspymgr.core import getSuperManager, User, UserType

app = None

def user_confirm_order(user):
    """用户成功下单时通知管理员
    """
    admin = getSuperManager()
    content = "用户(id: {}, email: {}) 于 {} 成功下单，请确认。".format(
        user.id, user.email, datetime.now())

    app.m_emailManager.add_email(
        to=admin.email,
        subject="用户下单",
        content=content,
        type="user confirm order")

from sspymgr.sscontroller import Account, AccountFlow

def registerApi(api):
    @api.route_admin('/product/getAll', methods=['POST'])
    def get_admin_products():
        prodcuts = Product.query.all()
        return jsonify({
            'status': 'success',
            'products': [p.to_dict() for p in prodcuts]
        })

    @api.route_admin('/product/get', methods=['GET', 'POST'])
    def admin_getProduct():
        id = request.form.get('id', -1)
        product = Product.query.filter_by(id=id).first()
        if product is None:
            return jsonify({'status': 'fail'})

        qrcodes = ProductQrcode.query.filter_by(pid=id).all()

        return jsonify({
            'status': 'success',
            'product': product.to_dict(),
            'qrcodes': [qr.to_dict() for qr in qrcodes],
        })

    @api.route_admin('/product/update', methods=['POST'])
    def admin_updateProduct():
        """更新或添加 Product
        """
        id = request.form.get('id', None)
        oldOne = Product.query.filter_by(id = id).first()
        if oldOne is None:
            oldOne = Product(price=5, duration=1296000,buffer_period=86400, flow=40960)
            app.m_db.session.add(oldOne)
            # return jsonify( {'status': 'fail'})
        oldOne.price = request.form.get('price', oldOne.price)
        oldOne.duration = request.form.get('duration', oldOne.duration)
        oldOne.flow = request.form.get('flow', oldOne.flow)
        oldOne.buffer_period = request.form.get('bufferPeriod', oldOne.buffer_period)
        # oldOne.enable = request.form.get('enable', oldOne.enable)
        enable = request.form.get('enable', None)
        if enable is not None:
            oldOne.enable = (enable == 'true')
        app.m_db.session.commit()
        return jsonify({'status': 'success'})

    @api.route_admin('/productQrcode/update', methods=['POST'])
    def admin_updateProductQrcode():
        # print(request.form)
        id = request.form.get('id')
        qrcode = ProductQrcode.query.filter_by(id=id).first()
        if qrcode is None:
            pid = request.form.get('productId', -1)
            if pid == -1:
                return jsonify({'status': 'fail'})
            qrcode = ProductQrcode(pid = pid, category='alipay')
            app.m_db.session.add(qrcode)
        qrcode.category = request.form.get('category', qrcode.category)
        qrcode.qrcode = request.form.get('qrcode')
        qrcode.desc = request.form.get('desc', qrcode.desc)
        qrcode.path = request.form.get('path', qrcode.path)
        qrcode.remark = request.form.get('remark', qrcode.remark)
        
        app.m_db.session.commit()
        return jsonify({'status': 'success'})

    @api.route_admin('/productQrcode/delete', methods=['POST'])
    def admin_deleteProductQrcode():
        id = request.form.get('id', -1)
        if id == -1:
            return jsonify({'status': 'fail'})
        qrcode = ProductQrcode.query.filter_by(id=id).delete()
        return jsonify({'status': 'success'})

    @api.route_admin('/orderHistory/get', methods=['POST'])
    def get_admin_order():
        page, per_page = api.getPageArgs()
        
        orders = Order.query.paginate(page=page, per_page=per_page)

        return jsonify({
            'status': 'success',
            'orders': orders
        })
    
    @api.route_admin('/orderHistory/getAll', methods=['GET', 'POST'])
    def admin_getAllOrderHistory():
        orders = Order.query.all()

        return jsonify({
            'status': 'success',
            'orders': [o.to_dict() for o in orders]
        })

    @api.route_user('/product/canPlace', methods=['GET', 'POST'])
    def can_user_place_order():
        """check whether user is able to place order
        """
        uid = session.get('userid')
        result = {'status': 'success'}
        account = Account.query.filter_by(userId=uid).first()
        if account is None:
            result['status'] = 'fail'
            result['canPlaceOrder'] = False
            return result
        now = datetime.now()
        next_time = account.expire - timedelta(days=1)
        if next_time > now:
            result['canPlaceOrder'] = False
            result['expire'] = formatTime(account.expire)
            result['nextPlaceOrderTime'] = formatTime(next_time)
        else:
            result['canPlaceOrder'] = True

        return jsonify(result)

    @api.route_user('/product/getAll', methods=['GET', 'POST'])
    def user_getAllProducts():
        """返回所有可用的 producte
        """
        products = Product.query.all()
        np = []
        for product in products:
            if product.enable:
                p = product.to_dict()
                p['alipay'] = ProductQrcode.query.filter_by(pid=product.id, category='alipay').count() > 0
                p['wechatpay'] = ProductQrcode.query.filter_by(pid=product.id, category='wechatpay').count() > 0
                np.append(p)

        return jsonify(np)

    @api.route_user('/product/getDetail', methods=['GET', 'POST'])
    def user_getProductDetail():
        """获取产品的详细信息
        """
        pid = request.form.get('product', -1)
        state = {'status': 'success'}
        product = Product.query.filter_by(id=pid).first()
        if pid == -1 or product is None:
            state['status'] = 'fail'
            return jsonify(state)

        p = product.to_dict()
        state['product'] = p
        qrcodes = ProductQrcode.query.filter_by(pid=pid).all()
        state['qrcodes'] = [qr.to_dict() for qr in qrcodes]
        return jsonify(state)

    @api.route_user('/product/confirm', methods=['POST'])
    def user_confirmOrder():
        """用户点击确认下单后的操作
        """
        state = {'status': 'success'}
        pid = request.form.get('product', -1)
        product = Product.query.filter_by(id=pid).first()
        if product is None:
            state['status'] = 'fail'
            return jsonify(state)

        comment = request.form.get('comment', '')
        uid = session['userid']
        flow = product.flow * 1024 * 1024  # product flow is stored in unit mb, convert it in byte

        # order history
        Order.place_order(uid, product, comment=comment)
        account = app.m_sscontroller.get_account_by_uid(uid)
        if account is None:
            state['status'] = 'fail'
        else:
            account.clear_flow()
            account.totalFlow = flow
            app.m_sscontroller.update_validity(account, seconds=product.duration, from_today=True)
            user = User.query.filter_by(id=uid).first()
            user.type = User.TypeActive
            user_confirm_order(user)
            
        app.m_db.session.commit()
        return jsonify(state)

    @api.route_user('/orderHistory/get', methods=['GET', 'POST'])
    def user_getOrderHistory():
        uid = session.get('userid')
        orders = Order.query.filter_by(uid=uid).all()
        order = [o.to_dict() for o in orders]
        return jsonify(order)

    logger.debug("api Registered")

from sspymgr import Manager
def init(iapp: Manager):
    global app
    app = iapp
    app.m_events.on('beforeRegisterApi', registerApi)
    init_product_type()
    logger.debug("inited")


# -*- coding: utf-8 -*-
"""Description: There are two main purpose of this module:
1. provide sqlalchemy as DB to other sub modules so they can define 
their database models even the database instance is not created yet.
2. provide a function called ``createDatabase`` to attach database to 
sqlalchemy, so it can get access to these databases (which would be sqlite
or mysql database).

Author: BriFuture

Modified: 2019/03/10 19:56
"""
from .path_helper import DATA_DIR

from os.path import abspath, join
def getDatabaseLoc(name: str):
    if ".db" not in name:
        name = "{}.db".format(name)
    return abspath( join( DATA_DIR, name ) )

"""
from . import app

app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///%s' % SQLITE_DATABASE_LOC
# app.config['SQLALCHEMY_COMMIT_ON_TEARDOWN'] = True
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True

from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy( app )
"""


### for pagenate

from sqlalchemy.orm import  Query
import math

def paginate(self, page=None, per_page=None, to_dict=True):
    """
    分页函数
    :param self:
    :param page:
    :param per_page:
    :return:
    """
    if page is None:
        page = 1

    if per_page is None:
        per_page = 20

    items = self.limit(per_page).offset((page - 1) * per_page).all()

    if not items and page != 1:
        return {'total': 0, 'page': page, 'error': 'no such items'}
        
    if page == 1 and len(items) < per_page:
        total = len(items)
    else:
        total = self.order_by(None).count()
    
    if to_dict:
        ditems = [item.to_dict() for item in items]
    else:
        ditems = items

    return {
        'page': page, 
        'per_page': per_page, 
        'total': total, 
        'items': ditems
    }
    # return Pagination(self, page, per_page, total, items)


from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy as sqlalchemy

Model = declarative_base()
sqlalchemy.Model = Model
# Database Proxy
DB = sqlalchemy
def createDatabase(name = None):
    """create instance of sqlalchemy Database
    """
    if name is None:
        name = 'database.db'
    if '.db' not in name:
        name = "{}.db".format(name)
        
    engine = sqlalchemy.create_engine('sqlite:///{}'.format(getDatabaseLoc(name)))
    session = orm.scoped_session(orm.sessionmaker(bind = engine))
    sqlalchemy.engine = engine
    Model.metadata.bind = engine
    Model.query = session.query_property()
    sqlalchemy.session = session
    sqlalchemy.create_all = Model.metadata.create_all


    Query.paginate = paginate
    return sqlalchemy



# -*- coding: utf-8 -*-
"""Description: This module is used for communicate with shadowsocks server, it contains basic protocol 
which defines how to communicate with shadowsocks server. Also contains a controller class to provide 
convenient methods for other modules.

Author: BriFuture

Detail: 2019/03/17 12:31
"""

from sspymgr import createLogger
logger = createLogger('sscontroller')

def _ping():
    """Return a tuple that is made up of a message that need to be sent to shadowsocks server 
    and the expected response from shadowsocks server
    """
    return b'ping', b'pong'

def _add_port(port, password, method):
    """Return a tuple that contains 4 elements
    First:  is the message that should be sent to shadowsocks server;
    Second: is the expected response from shadowsocks server;
    Third:  is a bool type value which indicates whether it is account-add operation.
    Fourth: is the port passed in.
    """
    cmd = 'add: {{ "server_port": {:d}, "password": "{:s}", "method": "{:s}" }}'.format(port, password, method)
    return cmd.encode( 'ascii' ), b'ok', True, port

def _remove_port(port):
    """Return a tuple that contains 4 elements, the same as ``_add_port``
    """
    cmd = 'remove: {{ "server_port": {:d} }}'.format(port)
    return cmd.encode( 'ascii' ), b'ok', False, port

# import re
# STAT = re.compile( b'stat: {[^s]+}' )
import json
from queue import Queue
from time import sleep
from .database import Account, AccountFlow
import socket 
# from gevent import socket, spawn, joinall
# from gevent.pool import Pool

class SSAdapter( object ):
    """Socket layer operation, it contains a queue which contains commands (ping or add_port or remove_port) that wait to be sent, 
    Pass a stater callback function as constructor parameter, and it will be called every time when shadowsocks statistics comes.

    Account States: created -> adding -> added -> invalid -> removed -> recreate
    accounts is a dict, contains port and flow.
    ..python::
        accounts = { 
            45000: Account, 
            45001: Account
        }
    """
    StateCreated = "created"
    StateAdding = "adding"
    StateAdded = "added"
    StateRemoving = "removing"
    StateRemoved  = "removed"
    StateRecreate = "recreate"
    ValidStates = ("adding", "added")
    InvalidStates = ("removed", "removing")
    MaxStatCount = 15

    def __init__( self, stater ):
        """``stater`` is the callback function which would be triggered every time shadowsocks server 
        send statistic messages.
        """
        # key is port: int, value is Account object
        self.accounts = dict()
        self._db = None
        self._stater = stater
        self._queue = Queue()
        self._stat_count = 0

    def conn_ser( self, server ):
        """Connect to shadowsocks server, ``server`` can be either a tuple for example ('127.0.0.1', 6060), either a str which is the location of unix sock file

        Note:
            unix sock file mode is not usable currently
        """
        # create a socket object, generally no exception here
        if type( server ) is tuple:
            self._socket = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )  ## UDP connection
        else:
            self._socket = socket.socket( socket.AF_UNIX, socket.SOCK_DGRAM )

        self._socket.settimeout( 0 )  # never timeout
        self._socket.setblocking( False ) # none block
        try:
            self._socket.connect( server )
        except OSError as e:
            logger.critical("Error when connecting to shadowsocks server: {}".format(e))

        # self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.ping()

    def ping( self ):
        """Tell the adapter to check whether shadowsocks server is online.
        """
        cmd = _ping()
        self._queue.put( cmd )

    def add_port( self, account: Account ):
        """Add an shadowsocks account with ``port``, ``password`` and ``method``
        accounts status process: created -> added
        """
        if account.port in self.accounts:
            self.accounts[account.port].status = self.StateRecreate
        else:
            self.accounts[account.port] = account
            account.status = self.StateCreated
        if account.status in self.ValidStates:
            # logger.info("Account {} is added, no need to duplicate add_port.".format(account))
            return
        cmd = _add_port(account.port, account.password, account.method)
        self._queue.put( cmd )
        logger.info("port is adding : {}".format(cmd))
    
    def remove_port( self, account: Account ):
        """Remove ``port`` from shadowsocks server, 
        accounts status process: removing => removed
        """
        if account.port not in self.accounts:
            self.accounts[account.port] = account
            
        account.totalFlow = 0
        account.clear_flow()
        if account.status in self.InvalidStates:
            logger.info("Account {} not valid, no need to remove_port.".format(account))
            return
        account.status = self.StateRemoving
        cmd = _remove_port(account.port)
        self._queue.put( cmd )
        logger.info("Port is removing : {}".format(cmd))
    
    def update_port( self, account: Account):
        """Convenient function for update ``password`` and ``method`` of account in shadowsocks server.
        """
        self.remove_port( account )
        self.add_port( account )

    def is_port_valid( self, port: int ):
        """Check if a port has been activated by the shadowsocks server
        """
        if port in self.accounts:
            return self.accounts[port].status == self.StateAdded
        return False

    def execute(self):
        """1. Check remaining commands in command queue and send them, 2. Recv message coming from shadowsocks server 
        """
        cmd = None
        if not self._queue.empty():
            cmd = self._queue.get()
            self._socket.send( cmd[0] )
            logger.debug("[sending]: {}".format(cmd))
            sleep(0.01)
        rmsg = None
        try:
            rmsg = self._socket.recv( 2048 )
            self._stat_count = 0
        except Exception as exc:
            if type(exc) != BlockingIOError:
                logger.warn("cmd executing error: {}".format(exc) )
        finally:
            self._stat_count += 1
            if( self._stat_count == self.MaxStatCount ):
                # avoid long time not recved stats information
                # logger.debug("SSServer not response for a long time")
                self._process_stats({})
                self._stat_count = 0
                
        if rmsg is None:
            return
        isstat = rmsg[:4] == b'stat'
        if isstat:
            stats = rmsg[5:]
            stats = json.loads( stats )
            self._process_stats(stats)
        else:
            if cmd is None:
                logger.info("Recv unrecognized message: {}".format(rmsg))
            else:
                self._process_response(rmsg, cmd)
                logger.debug("resp: {} cmd: {}".format(rmsg, cmd))


    def _process_response(self, rmsg, cmd):
        if rmsg != cmd[1]:
            logger.debug('recv msg: {} did not matchs expected: {}'.format(rmsg, cmd[1]))
            return

        if len(cmd) < 3:
            logger.info( 'Recved Pong from server' )
            return

        add = cmd[ 2 ]
        port = cmd[ 3 ]
        account = Account.query.filter_by(port = port).first()
        # account = self.accounts[port]
        account.status = self.StateAdded if add else self.StateRemoved
        account.used_flow = self.accounts[port].used_flow
        self._db.session.commit()
        self.accounts[port] = account
        logger.info( 'port {} is added: {}'.format(account, add) )

    def _process_stats(self, stats):
        dbaccounts = Account.query.all()
        accountFlow = AccountFlow.query.all()
        now = datetime.now()
        for account in dbaccounts:
            # attach accountFlow object to account
            for flow in accountFlow:
                if account.port == flow.port:
                    account.account_flow = flow
                    break
            if str(account.port) in stats:
                
                account.used_flow = stats[ str(account.port) ]
            else:
                account.used_flow = 0
            # account.used_flow = self.accounts[account.port].used_flow
            self.accounts[account.port] = account
        self._stater(self.accounts)


from datetime import datetime, timedelta

GB = 1024 * 1024 *1024 # unit: bytes
DEFAULT_METHOD = "aes-256-cfb"

class SSController(object):
    """Provide advanced api for manipulating shadowsocks server.
    """

    def __init__(self, db):
        """Start running the adapter that corresponds to communicate with shadowsocks server
        """
        self._db = db
        self._stats = None
        from . import ssAddr
        adapter = SSAdapter( self.process_stats )
        adapter._db = db
        adapter.conn_ser( server = ssAddr() )
        self._adapter = adapter
        self._start()

    def _start(self):
        # append account to shadowsocks server
        dbaccounts = Account.query.all()
        accountFlow = AccountFlow.query.all()
        now = datetime.now()
        for acc in dbaccounts:
            # attach accountFlow object to account
            for flow in accountFlow:
                if acc.port == flow.port:
                    acc.account_flow = flow
                    break
            # acc.status = SSAdapter.StateCreated
            acc.used_flow = 0
            # add port for this account no matter what state the account is (valid or invalid), 
            self._adapter.add_port(acc)

        self._db.session.commit()

    def setStats(self, stats):
        self._stats = stats

    def process_stats(self, accounts: dict):
        # if stats callback function is set, call it
        if self._stats is not None:
            self._stats({port: account.used_flow for port, account in accounts.items()})
        now = datetime.now()
        for port, account in accounts.items():
            # iterate all accounts stored in database records
            if account.status in SSAdapter.ValidStates and account.used_flow != 0:
                account.account_flow.updateTime = datetime.now()
                account.account_flow.flow += account.used_flow
                account.used_flow = 0
            if now > account.expire or not account.is_flow_remaining():
                # account is invalid now
                self._adapter.remove_port(account)
            elif account.status in SSAdapter.StateRemoved:
                self._adapter.add_port(account)
        self._db.session.commit()


    def add_account(self, port, password, flow = GB, **kwargs):
        """add a non-existing account to shadowsocks server, for existing accounts just 
        activate them will be fine
        @params 
            port: int = None
            password: str = None
            flow in the unit of byte,
            expire 
        @returns bool True if successfully added
        """
        if port in self._adapter.accounts:
            # Account already exists
            return False 
        # create account and record
        account = Account(port=port, password=password,
            # default flow: 1G
            totalFlow=flow,
            userId=kwargs.get("userId"),
            expire=kwargs.get("expire"),
            method = kwargs.get("method", DEFAULT_METHOD)
        )
        now = datetime.now()
        flow = AccountFlow(port=port, updateTime=now, checkTime=now, 
            autoBanTime = kwargs.get("expire"))
        account.account_flow = flow
        self._adapter.add_port(account)

        self._db.session.add(account)
        self._db.session.add(flow)
        self._db.session.commit()
        return True

    def del_account(self, port, **kwargs):
        """Delete an account in shadowsocks server, wont affect records in database
        """
        account = Account.query.filter_by(port=port).first()
        self._adapter.remove_port(account)
        self._db.session.commit()

    def get_account_by_id(self, id):
        """Return the account object if found, or None
        """
        # for port, account in self._adapter.accounts.items():
        #     if account.id == id:
        #         return account
        # return None
        account = Account.query.filter_by(id=id).first()
        if account is not None:
            account_flow = AccountFlow.query.filter_by(port = account.port).first()
            account.account_flow = account_flow
        return account

    def get_account_by_uid(self, uid) -> Account:
        """Return the account object if found, or None
        """
        # for port, account in self._adapter.accounts.items():
        #     if account.userId == uid:
        #         return account
        # return None
        account = Account.query.filter_by(userId=uid).first()
        if account is not None:
            account_flow = AccountFlow.query.filter_by(port = account.port).first()
            account.account_flow = account_flow
        return account

    def get_account_by_port(self, port) -> Account:
        """Return the account object if found, or None
        """
        # account = None
        # if port in self._adapter.accounts:
        #     account = self._adapter.accounts[port]
        # return account
        account = Account.query.filter_by(port=port).first()
        if account is not None:
            account_flow = AccountFlow.query.filter_by(port = account.port).first()
            account.account_flow = account_flow
        return account

    def update_totalflow(self, account: Account, bytes_flow: int):
        """Clear the flow usage on account, and update only ``totalFlow`` property of the account
        Note: ``db.session.commit`` has to be invokde by the caller
        """
        account.totalFlow = bytes_flow
        account.clear_flow()
        if not self._adapter.is_port_valid(account.port):
            self._adapter.add_port(account)
        return True

    def update_validity(self, account: Account, seconds: int, from_today=False):
        """Update validity property of an account 
        Note: ``db.session.commit`` has to be invokde by the caller
        @Return True indicate the port will be added, False removed
        """
        td = timedelta(seconds=seconds)
        now = datetime.now()
        if from_today:
            account.expire = now
        account.expire += td

        if account.expire > now:
            self._adapter.add_port(account)
            return True
        else:
            self._adapter.remove_port(account)
            return False

    def update_password(self, password, port=0, uid=None):
        """@deprecated Use update_password_by_port or update_password_by_uid instead.
        Update password of shadowsocks account. If ``uid`` is set, it's preferred be used to match the accounts;
        If ``port`` and ``uid`` are both set, it will check whther passed port matches the port in database record.
        Ohterwise, no action will be applied, False returned
        @returns True for success, False for failure
        """
        if uid is None:
            account = Account.query.filter_by(port=port).first()
        else:
            account = Account.query.filter_by(userId=uid).first()
            if port != 0 and account.port != port:
                return False
        if not account:  # in case account not found
            return False
        self._adapter.update_port(account)
        account.password = password
        self._db.session.commit()
        return True

    def update_password_by_port(self, password, port):
        """Update password of account(port identified)
        """
        account = Account.query.filter_by(port=port).first()
        return self._update_password(account, password)

    def update_password_by_uid(self, password, uid):
        """Update password of account(uid identified)
        """
        account = Account.query.filter_by(userId=uid).first()
        return self._update_password(account, password)

    def _update_password(self, account, password):
        if not account:  # in case account not found
            return False
        self._adapter.update_port(account)
        account.password = password
        self._db.session.commit()
        return True

    def execute(self):
        """send command and recieve message
        """
        self._adapter.execute()

    def list_account(self, atype='all'):
        """
        @ param atype
            'all'   list of all accounts
            'valid' list of all valid accounts
            'invalid' list of all invalid accounts
        @returns
            a list with all account
        """
        dbaccounts = Account.query.all()
        accountFlow = AccountFlow.query.all()
        for acc in dbaccounts:
            for flow in accountFlow:
                if acc.port == flow.port:
                    acc._flow = flow
                    break
        return dbaccounts

    def renew_account(self, flow, duration, port=None, uid=None):
        """@deprecated Use ``update_flow_validity`` instead
        Renew an account with total flow and validity
        @param 
            flow: in unit byte
            duration: timedelta
        """
        if port:
            acc = Account.query.filter_by(port=port).first()
        elif uid:
            acc = Account.query.filter_by(userId=uid).first()
        else:
            return False

        # it will cause more time instead of less
        date = datetime.now()
        if date < acc.expire:
            date = acc.expire
        expire = duration + date
        self._update_account(acc, flow = flow, expire = expire)
        self._db.session.commit()
        return True

    def _update_account(self, acc, expire: datetime = None, flow = GB):
        """@param acc: account
        """
        accFlow = AccountFlow.query.filter_by(port=acc.port).first()
        # A problem here is that user may lose some flow if last order is still active
        accFlow.flow = 0
        acc.totalFlow = flow
        if expire is not None:
            acc.expire = expire
        ## communicate with shadowsocks
        self._adapter.update_port(acc)


# -*- coding: utf-8 -*-
"""Description: This module is used for providing api interface for both admin and user frontend.
When providing apis with normal users, use decorator ``api.route_user``, 
while providing apis with administrator, use decorator ``api.route_admin``
It also provides a convinient function getPageArgs for requesting paged information.

Author: BriFuture

Modified: 2019/03/10 19:33
"""

from flask import Blueprint, request

api = Blueprint('api', __name__)

from datetime import timedelta, datetime
from functools import wraps
from flask import session, jsonify


def identify_user(func):
    """judge whether a user has signed in or not, 
    if he has signed in and he is administrator, redirect the page to admin_index
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if 'user' not in session:
            return jsonify({'status': 'Bad Request'})
        return func(*args, **kwargs)

    return wrapper


def user_route(rule, methods):
    def __user_route(func):
        @api.route('/user{}'.format(rule), methods=methods)
        @wraps(func)
        def wrapper(*args, **kwargs):
            if 'user' not in session:
                return jsonify({'status': 'Bad Request'})
            return func(*args, **kwargs)

        return wrapper

    return __user_route


api.route_user = user_route

from . import isManager

def identify_admin(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not isManager():
            # no need to redirect to similar page because there may be no such page for admin
            return jsonify({'status': 'Bad Request'})
        return func(*args, **kwargs)

    return wrapper


def admin_route(rule, methods):
    def __admin_route(func):
        @api.route('/admin{}'.format(rule), methods=methods)
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not isManager():
                # no need to redirect to similar page because there may be no such page for admin
                return jsonify({'status': 'Bad Request'})
            return func(*args, **kwargs)

        return wrapper

    return __admin_route


api.route_admin = admin_route

from flask import request


def getPageArgs():
    page = request.form.get('page', 1)
    per_page = request.form.get('perPage', 20)
    try:
        page = int(page)
        per_page = int(per_page)
    except:
        page = 1
        per_page = 20
    return page, per_page


api.getPageArgs = getPageArgs



# -*- coding:utf-8 -*-
"""sspymgr: a website application for managing shadowsocks service 
"""

__author__ = 'BriFuture'
__version__ = '0.0.21'

from .manager import Manager
# models
from .models import DB

from .path_helper import createLogger, createLoggerFile
from .globalfuncs import isEmailMatched, getRandomCode, convertFlowToByte, formatTime, tr


# -*- coding: utf-8 -*-

"""Description: Used for control shadowsocks, for example: add account, del account, flow usage stats.
The method ``start_shadowsocks`` integrates shadowsocks into sspymgr and it will start a shadowsocks server
in a separate thread. ``ssAddr`` can be used to make sure that shadowsocks server and adapter communicate 
with the same address, either unix socks file, either socket.

Author: BriFuture

Date: 2019/03/19
"""

__version__ = "0.0.6"

from sspymgr import convertFlowToByte, getRandomCode, Manager, createLogger
from sspymgr.core import User, WebguiSetting, UserType
logger = createLogger('sscontroller', stream=False, logger_prefix="[Core SSController]")

from .database import Account, AccountFlow
from .adapter import SSController

import sys
def ssAddr(addr2str = False):
    """Get universal address for communicating with shadowsocks server
    """
    if sys.platform.startswith( 'linux' ):
        mgr_addr = '/var/run/sspymgr.sock'
    else:
        mgr_addr = ('127.0.0.1', 6001)

    # some problem in unix sock file, using tcp mode always currently
    mgr_addr = ('127.0.0.1', 6001)

    if addr2str:
        mgr_addr = '{}:{}'.format( mgr_addr[0], mgr_addr[1])
    return mgr_addr

from .run_ss import main as start_shadowsocks

app = None
def init(iapp: Manager):
    global app
    app = iapp
    app.m_sscontroller = SSController(app.m_db)
    app.m_events.on('beforeRegisterApi', register_routes)

from flask import request, jsonify, session

from sqlalchemy.sql import func

def availiable_port(db) -> int:
    port = db.session.query(func.max(Account.port)).scalar()

    if port is None:
        port = WebguiSetting.getSetting(key='shadowsock_port_range_start',\
            default_value=45001, default_type="Number").getTypedValue()
        db.session.commit()
    else:
        port += 1
    return port

from datetime import datetime, timedelta
def register_routes(api):
    """register rotues for sscontroller
    """
        
    @api.route_admin('/user/getDetail', methods=['GET', 'POST'])
    def admin_getUserDetail():
        userId = request.form.get('userId')
        user = User.query.filter_by(id=userId).first()
        state = {'status': 'success'}
        if not user:
            state['status'] = 'fail'
            return jsonify(state)
        state['user'] = user.to_dict()
        account = Account.query.filter_by(userId=userId).first()
        if not account:
            state['account'] = {}
            return jsonify(state)
        state['account'] = account.to_dict()
        accFlow = AccountFlow.query.filter_by(port=account.port).first()
        state['accountFlow'] = accFlow.to_dict()
        return jsonify(state)


    @api.route_admin('/account/getDetail', methods=['POST'])
    def get_admin_accountDetail():
        id = request.form.get('id', None)
        state = {'status': 'success'}
        if id is None:
            state['status'] = 'fail'
            return jsonify(state)
        acc = Account.query.filter_by(id=id).first()
        state['account'] = acc.to_dict()
        accFlow = AccountFlow.query.filter_by(port=acc.port).first()
        state['accountFlow'] = accFlow.to_dict()
        if acc.userId:
            user = User.query.filter_by(id=acc.userId).first()
            state['user'] = user.to_dict()
        else:
            state['user'] = {}
        
        address = "{}:{}@{}:{}".format(acc.method, acc.password, acc.server, acc.port)
        state['ssAddress'] = address
        return jsonify(state)

    @api.route_admin('/account_port', methods=['POST'])
    def on_get_account_port():
        port = availiable_port(app.m_db)
        return str(port)

    @api.route_admin('/account/get', methods=['POST'])
    def get_admin_account():
        accounts = app.m_sscontroller.list_account()
        # accounts = Account.query.filter().all()
        return jsonify({
            'status': 'success',
            'accounts': [a.to_dict() for a in accounts]
        })

    @api.route_admin('/accountAdd', methods=['POST'])
    def add_account():
        port = request.form.get('port')
        try:
            port = int(port)
        except Exception:
            port = availiable_port(app.m_db)  # should get a useable port

        password = request.form.get('password')

        flow = request.form.get('flow')  # in the unit GB
        try:
            flow = int(flow)
        except:
            flow = 10
        flow = convertFlowToByte(flow, 'G')  # in the unit byte

        duration = request.form.get('duration')
        try:
            duration = int(duration)
        except Exception:
            duration = 7
        if not port or not password or not flow or not duration:
            return 'Bad request'
        expire = timedelta(days=duration) + datetime.now()
        app.m_sscontroller.add_account(port=port, password=password, flow=flow, expire=expire)
        logger.info('successfully add account %d' % port)
        return 'success'
    
    @api.route_admin('/account/updatePassword', methods=['POST'])
    def on_account_update_password():
        newpasswd = request.form.get('password', None)
        port = request.form.get('port', None)
        if newpasswd is None or port is None:
            return jsonify('fail')
        port = int(port)
        if app.m_sscontroller.update_password_by_port(newpasswd, port):
            return jsonify('success')
        else:
            return jsonify('fail')


    @api.route_admin('/account/updateTotalFlow', methods=['POST'])
    def on_account_update_totalflow():
        port = request.form.get('port', None)
        flow = request.form.get('flow', None)  # here the flow is in unit of Gb
        unit = request.form.get('flowUnit', 1024 * 1024 * 1024)
        try:
            port = int(port)
            flow = int(flow)
            unit = int(unit)
        except Exception:
            return jsonify('fail')
        account = Account.query.filter_by(port=port).first()
        if account is None:
            return jsonify('fail')

        flow = flow * unit
        account.totalFlow = flow
        
        app.m_db.session.commit()
        return jsonify('success')


    @api.route_admin('/account/randomPassword', methods=['POST'])
    def on_account_random_password():
        port = request.form.get('port', None)
        result = {'status': 'fail'}
        if port is None:
            return jsonify(result)
        port = int(port)
        newpasswd = getRandomCode(8)
        if app.m_sscontroller.update_password_by_port(newpasswd, port):
            result['status'] = 'success'
            result['password'] = newpasswd
        else:
            result['status'] = 'fail'
        return jsonify(result)

    DAY_SENCONDS = 86400
    @api.route_admin('/account/updateExpiration', methods=['POST'])
    def on_account_update_expiration():
        result = {'status': 'fail'}
        port = request.form.get('port', None)
        days = request.form.get('days', None)
        try:
            port = int(port)
            days = int(days)
        except Exception:
            return jsonify(result)
        result['status'] = 'success'
        # acc = Account.query.filter_by(port=port).first()
        acc = app.m_sscontroller.get_account_by_port(port)
        if not acc:
            result['status'] = 'fail'
            return jsonify(result)
        seconds = days * DAY_SENCONDS
        added = app.m_sscontroller.update_validity(acc, seconds)
        user = User.query.filter_by(id = acc.userId).first()
        if added:
            if user is not None:
                user.type = UserType.ACTIVE.value
            if acc.totalFlow == 0:
                setting = WebguiSetting.getSetting("default_account_flow",\
                    default_value=1, default_type="Number")
                acc.totalFlow = setting.getTypedValue() * 1024 * 1024 * 1024
        else:
            if user is not None:
                user.type = UserType.INVALID.value
        app.m_db.session.commit()
        result['nexpire'] = acc.expire.timestamp()
        return jsonify(result)
    
    @api.route_user('/accountDetail', methods=['POST'])
    def user_accountDetail():
        uid = session['userid']
        account = Account.query.filter_by(userId=uid).first()
        result = {'status': 'fail'}
        if account is not None:
            accountFlow = AccountFlow.query.filter_by(port=account.port).first()
            account._flow = accountFlow
            server = request.host
            if ":" in server:
                server = server.split(':')[0]
            enabled = (account.status == 'added')
            account.server = account.server or server
            account.last_connected = accountFlow.updateTime

            # 用于生成二维码的地址
            # address = '-'.join(
            #     [account.server,
            #     str(account.port), account.password])
            address = "{}:{}@{}:{}".format(account.method, account.password, account.server, account.port)

            result['account'] = account.to_dict()
            result['ssAddress'] = address
            result['enabled'] = enabled
            result['status'] = 'success'
        return jsonify(result)


    @api.route_user('/account/updatePassword', methods=['POST'])
    def on_user_update_account_password():
        password = request.form.get('password')
        if not password:
            return jsonify('failed')
        uid = session['userid']
        return jsonify('success') if app.m_sscontroller.update_password_by_uid(password, uid) else jsonify('fail')
    
    logger.info("Api Registered.")



# -*- coding: utf-8 -*-
"""Description: This module may be the most significant part of the webserver. It contains many sub systems
such as Signin, Singup and Administrator management. At the very begging of this project, Chiness and English
annotations or text are mixed in many parts, so one of the jobs TODO is to rewrite all of them in English 
and then use ``gettext`` module to translate them into Chinses.

Author: BriFuture

Date: 2019/03/18 21:51
"""

from sspymgr import DB, createLogger
from sspymgr import getRandomCode, convertFlowToByte

from flask import session, request, jsonify
from datetime import datetime, timedelta
import hashlib

logger = createLogger(
    "core_user", stream=False, logger_prefix="[Plugin User]")


class User(DB.Model):
    """The life cycle of an webserver user is not clear now. So the user type may be
    confusing. TODO reconstruct this model
    """
    __tablename__ = 'user'
    id = DB.Column(DB.Integer, primary_key=True, autoincrement=True)
    type = DB.Column(DB.String(255))
    salt = DB.Column(DB.String(8), nullable=False)
    email = DB.Column(DB.String(255), nullable=False)
    group = DB.Column(DB.Integer, default=0)
    level = DB.Column(DB.Integer, default=1)
    comment = DB.Column(DB.String(255), default='')
    password = DB.Column(DB.String(255), nullable=False)
    telegram = DB.Column(DB.String(255))
    username = DB.Column(DB.String(255))
    lastLogin = DB.Column(DB.DateTime)
    createTime = DB.Column(DB.DateTime)
    resetPasswordId = DB.Column(DB.String(255))
    resetPasswordTime = DB.Column(DB.DateTime)

    TypeTrial  = 'trial'
    TypeActive = "active"
    TypeBanned = 'banned'
    TypeInvalid = 'invalid'
    TypeNotified = 'notified'
    TypeDefault = 'default'
    TypeTest = 'test'
    def __init__(self, **kwargs):
        super(User, self).__init__(**kwargs)
        # self.password = password
        self.salt = getRandomCode(8)
        self.password = self.__gen_passwd(self.password)
        self.createTime = datetime.now()

    def __repr__(self):
        return '<User %s %s %s>' % (self.email, self.salt, self.password)

    def __gen_passwd(self, password):
        """Just generate secure password, but not modify object's password attribute,
        parameter password can't be None or self.password, otherwise the method 
        which checks whether password is valid will return a Fake result
        """
        return hashlib.sha256((self.salt + password).encode()).hexdigest()

    def updatePassword(self, new_passwd):
        self.password = self.__gen_passwd(new_passwd)

    def is_password_valid(self, password):
        """check whether the password (will be encrypted) passed in is the same 
        as the encrypted password stored in the database records
        """
        password = str(password)
        password = self.__gen_passwd(password)
        return password == self.password

    def request_reset_passwd(self):
        self.resetPasswordId = getRandomCode(24)
        self.resetPasswordTime = datetime.now()

    def do_reset_passwd(self, passwdId, new_passwd):
        """reset password and clear requestId
        @returns
            True if succesfully reset, False otherwise
        """
        if not self.resetPasswordId or not self.resetPasswordTime:
            return False
        now = datetime.now()
        diff = now - self.resetPasswordTime

        if diff > timedelta(hours=3):
            # overtime, invalid
            return False

        if passwdId != self.resetPasswordId:
            return False

        self.password = self.__gen_passwd(new_passwd)
        self.resetPasswordId = ''
        return True

    def to_dict(self):
        lastLogin = self.lastLogin.timestamp(
        ) if self.lastLogin is not None else None
        di = {
            'id': self.id,
            'email': self.email,
            'type': self.type,
            'group': self.group,
            'comment': self.comment,
            'telegram': self.telegram,
            'username': self.username,
            'lastLogin': lastLogin,
            'createTime': self.createTime.timestamp(),
        }
        return di


class Group(DB.Model):
    __tablename__ = 'group'
    id = DB.Column(DB.Integer, primary_key=True, autoincrement=True)
    name = DB.Column(DB.DateTime)
    order = DB.Column(DB.String(255))
    comment = DB.Column(DB.DateTime)
    showNotice = DB.Column(DB.Integer, default=1)
    multiAccount = DB.Column(DB.Integer, default=0)


from enum import Enum, unique

# TODO create a clear life cycle for ``User`` and records it into a table
@unique
class UserType(Enum):
    BANNED = 'banned'
    TRIAL = 'trial'
    ACTIVE = 'active'
    INVALID = 'invalid'
    NOTIFIED = 'notified'
    DEFAULT = 'default'
    TEST = 'test'


# class UserStatus(DB.Model):
#     __tablename__ = 'UserStatus'
#     userId = DB.Column(DB.Integer, nullable=False)
#     webStatus = DB.Column(DB.String(64), default=UserType.DEFAULT.value)
#     emailStatus = DB.Column(DB.String(64))
#     ssStatus = DB.Column(DB.String(64))


def getCurrentUser() -> User:
    """Get currently logged in user data object
    """
    uid = session.get('userid')
    user = User.query.filter_by(id=uid).first()
    return user


def isManager(user=None) -> bool:
    """If user is set, check if user is manager;
    Else check it by session
    """
    if user is not None:
        return user.id == 1
    if 'userid' in session:
        return session.get('userid', -1) == 1
    return False


def getSuperManager() -> User:
    """Get the super manager.
    """
    return User.query.filter_by(id=1).first()


def _user_signin(user):
    """process user signin action, assume user is validated
    """
    session['user'] = user.email
    session['userid'] = user.id
    user.lastLogin = datetime.now()

from sspymgr import tr
def _send_checkcode_email(emailaddr, emailManager):
    """get check code when signup
    """
    rcode = getRandomCode()
    session['usercheckcode'] = rcode
    # 你好，你在 SSPY-MGR 中的注册验证码为:
    content = tr('Hello. Your checkcode in SSPY-MGR is:\n {}').format(rcode)
    emailManager.add_email(
        to=emailaddr,
        subject='SSPY SignUp',
        content=content,
        type="signup checkcode")


def _send_reset_link(host, email, code, emailManager):
    """generate reset password link
    """
    link = host + '/vhome#/resetPassword/' + code
    # url_for( 'home.on_resetPasswdWithId', code = code)
    
    content = tr("""Hi, you request to reset your password at SSPY-MGR.
    Here your reset password link for SSPY-MGR is: {}, please reset your password within 3 days.
    If you never request to reset password, please ignore this email.
    """).format(link)
    emailManager.add_email(
        to=email,
        subject="SSPY Reset Password Link",
        content=content,
        type="reset password")


MAX_CHECKCODE_ATTEMPT = 3
CHECKCODE_TIMEOUT = timedelta(minutes=10)

from sspymgr import isEmailMatched

def registerApi(api, app):
    from . import WebguiSetting

    @api.route('/home/signin', methods=['POST'])
    def on_home_signin():
        """Basic entry to signin for users who want to use sspymgr
        """
        # first of all, validate input data
        resp = {'status': 'success'}
        emailaddr = request.form.get('email', '')
        password = request.form.get('password', '')
        valid = True
        if (len(emailaddr) == 0):
            resp['email'] = 'Email is Empty'
            valid = False
        if (len(password) == 0):
            resp['password'] = 'Password is Empty'
            valid = False

        if valid is False:
            resp['status'] = 'fail'
            return jsonify(resp)
        # find correspoding user and if not found, raise error
        user = User.query.filter_by(email=emailaddr).first()
        if user is None:
            resp['status'] = 'fail'
            resp['email'] = 'Email is not valid'
            return jsonify(resp)

        if user.is_password_valid(password):
            # able to sign in
            _user_signin(user)
            session.permanent = request.form.get(
                'keep_signedin') == 'remember-me'
            resp['status'] = 'success'
            resp['level'] = 'admin' if isManager(user) else 'user'
        else:
            # unable to signin, password may be wrong
            resp['status'] = 'fail'
            resp['password'] = 'Password is wrong'
        return jsonify(resp)

    @api.route('/home/isSignedIn', methods=['POST'])
    def is_user_signedIn():
        resp = {'signedIn': False, 'level': 'user'}
        if 'userid' in session:
            resp['signedIn'] = True
            if isManager():
                resp['level'] = 'admin'

        return jsonify(resp)

    @api.route('/home/checkcode', methods=['POST'])
    def on_send_checkcode():
        resp = {'status': 'success'}
        email = request.form.get('email', '')
        if len(email) == 0:
            resp['status'] = 'fail'
            resp['desc'] = 'Email is empty'
            return jsonify(resp)

        exists = User.query.filter_by(email=email).count() > 0
        if not isEmailMatched(email) or exists:
            resp['status'] = 'fail'
            resp['desc'] = 'Wrong Email or Email already in use'
            return jsonify(resp)

        if 'checkcode_attempt' not in session:
            session['checkcode_attempt'] = 0
            session['checkcode_start'] = datetime.now().timestamp()

        singup_limit = WebguiSetting.getSetting('signup_remain_limit',\
            default_value=4, default_type="Number").getTypedValue()
        if singup_limit == 0:
            resp['status'] = 'fail'
            resp['desc'] = 'Sorry, no more users can signup.'
            return jsonify(resp)

        session['checkcode_attempt'] += 1

        if session['checkcode_attempt'] <= MAX_CHECKCODE_ATTEMPT:
            _send_checkcode_email(email, app.m_emailManager)
        else:
            start = datetime.fromtimestamp(session['checkcode_start'])
            now = datetime.now()
            if now - start > CHECKCODE_TIMEOUT:
                session['checkcode_attempt'] = 1
                _send_checkcode_email(email, app.m_emailManager)
            else:
                resp['status'] = 'fail'
                resp[
                    'desc'] = 'Please wait a few minutes before resend check code.'
        app.m_db.session.commit()
        return jsonify(resp)

    @api.route('/home/signup', methods=['POST'])
    def on_home_signup():
        resp = {'status': 'fail'}
        valid = True
        if (not 'usercheckcode' in session):
            resp['code'] = 'Get checkcode first.'
            valid = False

        email = request.form.get('email', '')
        checkcode = request.form.get('checkcode', '')
        password = request.form.get('password', '')
        if (len(email) == 0):
            resp['email'] = 'Email is Empty'
            valid = False
        if (len(checkcode) == 0):
            resp['code'] = 'CheckCode is Empty'
            valid = False
        if (len(password) == 0):
            resp['password'] = 'Password is Empty'
            valid = False
        if valid and not isEmailMatched(request.form['email']):
            resp['email'] = 'Wrong email or not able to sign up now'
            valid = False

        if valid is False:
            return jsonify(resp)

        user = User.query.filter_by(email=email).first()
        if user is not None:
            resp['status'] = 'fail'
            resp['email'] = 'Email has already in use'
            return jsonify(resp)

        if (session.get('usercheckcode').lower() == checkcode.lower()):
            email = email.strip()
            user = User(email=email, password=password)
            app.m_db.session.add(user)
            setting = WebguiSetting.getSetting(key='signup_remain_limit')
            setting.value = setting.getTypedValue() - 1
            app.m_db.session.commit()
            # add account into shadowsocks server, but there is a problem is that if not committed,
            # user.id will be None
            user = User.query.filter_by(email=email).first()
            app.m_events.trigger(
                "webuser_signup_success",
                dict(
                    userId=user.id,
                    flow=convertFlowToByte(1, 'G'),
                ))

            # checkcode will be removed after user have been added
            session.pop('usercheckcode')
            # auto signin
            resp['status'] = 'success'
            _user_signin(user)

            app.m_db.session.commit()
        return jsonify(resp)

    @api.route('/home/requestReset', methods=['POST'])
    def on_home_requestReset():
        state = {'status': 'success'}
        email = request.form.get('email', '')
        if (len(email) == 0 or not isEmailMatched(request.form['email'])):
            state['status'] = 'fail'
        user = User.query.filter_by(email=email).first()

        if user:
            user.request_reset_passwd()
            code = '%d_%s' % (user.id, user.resetPasswordId)
            _send_reset_link(request.host, email, code, app.m_emailManager)
            app.m_db.session.commit()
        else:
            state['status'] = 'fail'
        return jsonify(state)

    @api.route('/home/resetPassword', methods=['POST'])
    def on_home_doResetPasswdWithId():
        code = request.form.get('code', '')
        code = code.split('_')
        state = {'status': 'success'}
        if len(code) < 2:
            state['status'] = 'fail'
            state['hint'] = 'Wrong code'
            return jsonify(state)
        userId = code[0]
        resetId = code[1]
        user = User.query.filter_by(id=userId, resetPasswordId=resetId).first()
        newpw = request.form.get('newPassword', '')
        if user is None or len(newpw) == 0:
            state['status'] = 'fail'
            state['hint'] = 'Wrong Password'
            return jsonify(state)

        result = user.do_reset_passwd(resetId, newpw)
        app.m_db.session.commit()
        if not result:
            state['status'] = 'fail'
            state['hint'] = 'Your code seems not right or expired'
        return jsonify(state)

    # =====================  user route =====================
    @api.route_user('/signout', methods=['POST'])
    def user_signout():
        session.pop('user')
        session.pop('userid')
        return jsonify('success')

    @api.route_user('/updatePassword', methods=['POST'])
    def on_user_update_password():
        uid = session.get('userid')
        user = User.query.filter_by(id=uid).first()
        oldp = request.form.get('oldPassword')
        if user is None or not user.is_password_valid(oldp):
            return jsonify('fail')
        newp = request.form.get('newPassword')
        user.updatePassword(newp)
        app.m_db.session.commit()
        return jsonify('success')

    # ===============  admin route  =================
    @api.route_admin('/user/updateInfo', methods=['POST'])
    def admin_updateUserInfo():
        comment = request.form.get('comment')
        utype = request.form.get('type')
        uid = request.form.get('userId')
        user = User.query.filter_by(id=uid).first()
        if user is None:
            return jsonify('fail')
        user.comment = comment
        user.type = utype
        app.m_db.session.commit()
        return jsonify('success')

    @api.route_admin('/user/getAll', methods=['POST'])
    def get_admin_user():
        users = User.query.filter().all()
        return jsonify({
            'status': 'success',
            'users': [u.to_dict() for u in users]
        })

    logger.info("api registered.")


# -*- coding: utf-8 -*-
"""Description: This module controls the configuration class and configuration parse processing.
Command line arguments provided by user will be parsed within the configuration class ``Configuration``.
In order to get the same config object within other sub modules, there is a function called ``defaultConfig``
which returns singliton instance of Configuration.

Author: BriFuture
Modified: 2019/03/10 19:43
"""
import os
basedir = os.path.abspath( os.path.dirname( __file__) )
import yaml
from pathlib import Path
from .path_helper import CONFIG_PATH
from .globalfuncs import tr

class Configuration:
    """This class is used for integrate configuration file reading/writing and command line arguments parsing.
    DO NOT instance this class directly, use ``defaultConfig`` function instead
    """
    SECRET_KEY = os.environ.get( 'SECRET_KEY' ) or 'ITISAHARDGUESSSTRING'
    SQLALCHEMY_COMMIT_ON_TEARDOWN = True
    
    def __init__(self, *args, **kwargs):
        self.parse()
        pass
    
    def parse( self ):
        if not os.path.exists( CONFIG_PATH ):
            _write_default_config()

        ## part of arguments from command line will cover the configuration from config.file
        self.__parse_sys_argv()
        self.__check_configuration()

    def __parse_yaml(self, path: Path):
        with path.open( 'r', encoding='utf-8' ) as f:
            yamlObj = yaml.load( f )
            self.raw_config = yamlObj
            mgrconfig = yamlObj['sspymgr']
            self.email = yamlObj['email']
            self.debug = False
            if 'debugConfig' in mgrconfig:
                if mgrconfig['debugConfig']['enabled']:
                    config = mgrconfig['debugConfig']
                    self.debug = True
                else:
                    config = mgrconfig['config']
            else:
                config = mgrconfig['config']
            self.host = config['host']
            self.port = config['port']
            self.shadowsocks = config.get('shadowsocks', True)
            self.webserver   = config.get('webserver', True)

    def __parse_sys_argv(self):
        import argparse
        parser = argparse.ArgumentParser(description=tr('Run SSPYMGR. Support Python3.x only.'))

        self.__add_arguments(parser)

        args = parser.parse_args()
        
        self.database = "database"

        if args.config is not None:
            index = args.config.find(".yaml")
            try:
                p = Path(args.config)
                if index > -1:
                    self.database = args.config[:index]
                self.__parse_yaml(p)
            except:
                p = Path(CONFIG_PATH)
                self.__parse_yaml(p)
        else:
            p = Path(CONFIG_PATH)
            self.__parse_yaml(p)

            
        if args.webhost is not None:
            self.host = args.webhost
        if args.disable_ss:
            self.shadowsocks = False
        if args.only_ss:
            self.shadowsocks = True
            self.webserver = False


        self.stream_log = args.log_console
        # -w start web gui server
        # -s start shadowsocks server

    def __add_arguments(self, parser):
        parser.add_argument("--log_console", 
            help=tr("Logger output will send to stdout."), action="store_true")
        parser.add_argument("-c", "--config", 
            help=tr("Specify configuration file (absolute path), sspymgr will create a new database file whose name is the same as the configuration file, \
                for example, -c ~/.sspy-mgr/local.yaml will create a database named ~/.sspy-mgr/local.db"), nargs="?")

        webgroup = parser.add_argument_group('website', 'website options')
        webgroup.add_argument("--webhost", help=tr("Set IP address that web werver listens, \
            for example 0.0.0.0 for IPv4, :: for IPv6"), nargs="?")
        webgroup.add_argument("--webport", help=tr("Set port that Website listens, usually 80"), nargs="?")

        ssgroup = parser.add_argument_group('shadowsocks', tr('integrated shadowsocks opration'))
        ssgroup.add_argument("--disable-ss", help=tr("Enable backend shadowsocks server,\
            disable it when developing the program may be helpful"), 
            action="store_true")
        ssgroup.add_argument("--only-ss", help=tr("Only start backend shadowsocks server \
            with defined communication method (socket or unix sock file) "), 
            action="store_true")

    def __check_configuration(self):
        ## TODO check host valid

        pass

def _write_default_config():
    """only call this method when needed, or config file may be overrided
    """
    with open( CONFIG_PATH, 'w', encoding='utf-8' ) as f:
        default = """email:
  account: 'your@account.xx'
  password: 'your password'
  host: 'smtp host'
sspymgr:
  debugConfig: 
    enabled: false # true to override following config
    host: '127.0.0.1'
    port: 5050
    shadowsocks: false
  config:
    host: '0.0.0.0'
    port: 80
    shadowsocks: true
"""
        f.write( default )

__config = None
def defaultConfig() -> Configuration:
    """get single instance of the configuration 
    """
    global __config
    if __config is None:
        __config = Configuration()
    
    return __config


# -*- coding: utf-8 -*-

"""Description: Basic integrated entry for starting webserver and shadowsocks server in a main looped thread, 
It's possible to start only webserver or shadowsocks server seperately. But it needs some additional config 
in configuration file or with command arguments. SSPYMGR is not featured yet although it can work, so it won't be
updated for some time until I get some spare time to reconstruct it. 

Author: BriFuture

Modified: 2019/03/10 19:17
"""

__author__ = 'BriFuture'

from gevent import monkey
monkey.patch_all ()

from .configuration import defaultConfig
from .path_helper import createLogger
from .globalfuncs import initGettext
tr = initGettext()
logger = createLogger("sspymgr")
config = defaultConfig()

app = None
def run_forever():
    """Basic integrated entry for starting webserver and shadowsocks server in a main looped thread
    """
    # from 
    if config.stream_log:
        import sspymgr.path_helper as ph
        ph.log_to_console = True
    run_shadowsocks(config)
    if config.webserver:
        run_webserver()
    else:
        logger.info("Webserver sspymgr webserver not start. Only shadowsocks will running")

def run_shadowsocks(config):
    """Start shadowsocks through api provided by sscontroller sub module.
    """
    if config.shadowsocks or not config.webserver:
        from .sscontroller import start_shadowsocks
        wait = not config.webserver
        start_shadowsocks(wait=wait)
    else:
        logger.info("shadowsocks won't start.")

from .manager import Manager


def run_webserver():
    """Start webserver application, the application is an instance of Mananger which is a subclass of Flask
    Currently it will only show debug info when running under debug mode, but it's a little annoying that the 
    templates won't be auto reloaded if the application is not running under debug mode.
    Or maybe it's more resanonable if this function is a method of Manager. But now it will keep unchanged.
    """
    global app
    app = Manager( __name__ )
    from .core import init as core_init
    core_init(app)
    app.start()

    from gevent import pywsgi
    weblogger = createLogger("webwsgi")

    if config.debug:
        from werkzeug.debug import DebuggedApplication
        app.debug = True
        app.jinja_env.auto_reload = True
        app.config['TEMPLATES_AUTO_RELOAD'] = True
        dapp = DebuggedApplication( app, evalex= True)
        server = pywsgi.WSGIServer( ( config.host, config.port ), 
            dapp, log = weblogger )
        logger.info( 'starting web server...')
        server.serve_forever()
    else:
        app.debug = False
        server = pywsgi.WSGIServer( ( config.host, config.port ), 
            app, log = weblogger )
        logger.info('starting web server...')
        server.serve_forever()




# -*- coding: utf-8 -*-
"""Description: Used for integrating all necessary objects and providing more convenient access 
to those objects.

Author: BriFuture

Modified: 2019/03/10 19:48
"""
from flask import Flask

from .dbsessions import replaceSessionInterface

from .configuration import defaultConfig
from .models import createDatabase
from ._plugin import load_plugins

import threading
def run_threaded(job_func):
    job_thread = threading.Thread(target=job_func)
    job_thread.start()


class Manager(Flask):
    """Sub class of Flask, it would support better interface 
    .. Properties:: global variables
        ``m_config``: Config which is set by configuration file or sys.argv
        ``m_db``: sqlalchemy, database reference
        ``m_events``: plugin.Events, events control and trigger
        ``m_tasker``: plugin.BackTask, run background tasks such as shadowsocks flow stats
        ``m_sscontroller``: sscontroller which is used to communicate with shadowsocks server
        ``m_emailManager``: emailManager which is used to send and record email
    """
    m_config = None
    m_db = None
    m_events = None
    # EmailManager
    m_emailManager = None
    # SSController
    m_ssctronller = None

    def __init__(self, *args, **kwargs):
        """initialize all variables and databases.
        return a Flask instance app.
        """
        super().__init__(*args, **kwargs)
        self.__prepare()
        replaceSessionInterface(self)

    def start(self):
        """make everythin prepared and then run the manager
        """
        self.m_events.on("task_schedule", self.scheduleTasks)
        load_plugins(self)
        
        # introduce plugin system, plugins should be inited before routes
        self.m_events.trigger('beforeCreateDb', eventArgs = self.m_db)
        self.m_db.create_all()
        self.m_events.trigger('afterCreateDb', eventArgs = self.m_db)
        
        from .sscontroller import init as init_ss
        init_ss(self)
        
        self._init_routes()
        self.m_tasker.start()

    def scheduleTasks(self, schedule):
        schedule.every().seconds.do(run_threaded, self.m_emailManager.checkRemain)
        schedule.every().seconds.do(run_threaded, self.m_sscontroller.execute)
        self.logger.debug("Email and SSController is scheduled")
        

    def __prepare(self):
        from .path_helper import createLogger
        from ._plugin import Events, BackTask
        self.m_config = defaultConfig()
        self.m_db = createDatabase(self.m_config.database)
        self.logger = createLogger("sspymgr")

        self.m_events = Events()
        self.m_tasker = BackTask(self.m_events)

    def _init_routes(self):
        # from .routes.vuefront import vuefront

        @self.teardown_request
        def teardownd_database(resp):
            self.m_db.session.close()
            return resp

        from .core import api
        # register core plugins and other plugins
        self.m_events.trigger('beforeRegisterApi', eventArgs = api )

        # add api routes which will start as `/api`
        self.register_blueprint( api, url_prefix='/api')
    


# -*- coding: utf-8 -*-
"""Description: Provide general functions that may be used in other sub modules. Note: Most functions are made
for convert bytes flow or time, most of them may seem not so useful because the rendering process now moved 
into client side and the Frontend HTML UI is now driven by VUE. 

## TODO clear those functions that are no longer used, integrated path_helper with this module because they do
the similar things.

Author: BriFuture

Date: 2019/03/19 23:07
"""
import hashlib, random, string
code_lib = string.ascii_letters + string.digits

import re
EMAIL_REGEX = re.compile(r'[^@]+@[^@]+\.[^@]+')

def isEmailMatched(email: str) -> bool:
    return EMAIL_REGEX.match(email)

def getRandomCode( bit = 6 ) -> str:
    return ''.join( random.SystemRandom().choices( code_lib, k = bit ) )

def convertFlowToByte( flow: int, unit: string ):
    """
    convert flow into bytes
    @param unit: UpperCase, could be G, M, K
    """
    if unit == 'GB' or unit == 'G':
        return flow * 1024 * 1024 * 1024
    elif unit == 'MB' or unit == 'M':
        return flow * 1024 * 1024
    elif unit == 'KB' or unit == 'K':
        return flow * 1024
    
    return flow

def convertByteToFlow( bytes: int ):
    # if bytes < 1024:
    #     return '%d b' % bytes
    # elif bytes < 1024 * 1024:
    #     return '%d Kb' % bytes / 1024
    mb = bytes / 1024 / 1024
    if mb < 1024:
        return '%d Mb' % mb

    gb, mb = divmod( mb, 1024 )
    if mb != 0:
        return '%d Gb %d Mb' % ( gb, mb )
    else:
        return '%d Gb' % gb

from datetime import datetime
def formatTime( dt: datetime ):
    """given datetime object, return string format, 
    second will be hide
    """
    return dt.strftime( '%Y-%m-%d %H:%M' )

def second2str( seconds: int ):
    days, remains = divmod( seconds, 86400 )
    hours, remains = divmod( remains, 3600 )
    minutes, remains = divmod( remains, 60 )
    str = ''
    if days != 0:
        str += "%d days" % days
    if hours != 0:
        str += "%d hours" % hours
    if minutes != 0:
        str += "%d minutes" % minutes
    return str

def mbflow2str( flow: int ):
    """unit of flow is MB
    """
    str = None
    if flow < 1024:
        return '%d Mb' % flow

    gb, mb = divmod( flow, 1024 )
    if mb != 0:
        return '%d Gb %d Mb' % ( gb, mb )
    else:
        return '%d Gb' % gb

from pathlib import Path

LANGUAGE_DIR = (Path(__file__).parent / "locale").resolve()

import gettext
def initGettext(domain="sspymgr") -> gettext.gettext:
    """make some configurations on gettext module 
    for convenient internationalization
    """
    gettext.bindtextdomain(domain, LANGUAGE_DIR)
    gettext.textdomain(domain)
    gettext.find(domain, "locale", languages=["zh_CN", "en_US"])
    return gettext.gettext

tr = initGettext()


# -*- coding: utf-8 -*-
"""Description: This module is used to provide email feature to all
other sub modules so that they can send email by a simple api.

Author: BriFuture

Modified: 2019/03/17 12:35
"""


__version__ = '0.1.2'

from sspymgr import DB, createLogger
from datetime import datetime
logger = createLogger('email', stream=False)

class Email( DB.Model ):
    __tablename__ = 'email'
    id = DB.Column( DB.Integer, primary_key = True, autoincrement = True )
    to = DB.Column( DB.String( 255 ), nullable=False )
    content = DB.Column( DB.Text )
    time = DB.Column( DB.DateTime, default=datetime.now )
    subject = DB.Column( DB.String( 255 ) )
    ip = DB.Column( DB.String( 255 ) )
    type = DB.Column( DB.String( 255 ) )
    remark = DB.Column( DB.String( 255 ) )
    session = DB.Column( DB.String( 255 ) )

    def to_dict(self):
        di = {
            'id': self.id,
            'to': self.to,
            'content': self.content,
            'time': self.time.timestamp(),
            'subject': self.subject,
            'ip': self.ip,
            'type': self.type,
            'remark': self.remark,
            'session': self.session,            
        }
        return di

    def __repr__(self):
        return "<Email to: {}, subject: {}, content: {}, time: {}>"\
            .format(self.to, self.subject, self.content, self.time)

from queue import Queue
import smtplib
import yaml
from email import encoders
from email.header import Header
from email.mime.text import MIMEText
from email.utils import parseaddr, formataddr


class MailSender(object):
    """``EmailSender`` is a wrapper of ``email`` module
    """
    def __init__(self, host, account, password, port, *args, **kwargs):
        self._host = host
        self._account = account
        self._password = password
        self._port = port
        self.server = smtplib.SMTP_SSL(self._host, self._port)

    def reset_server(self):
        self.server = smtplib.SMTP_SSL(self._host, self._port)

    def send(self, email: Email):
        msg = MIMEText(email.content, 'html', 'utf-8')
        msg['Subject'] = Header(email.subject, 'utf-8').encode()
        msg['From'] = self.__format_addr('SSPY-Mgr <{}>'.format(self._account))
        msg['To'] = self.__format_addr('Receiver <{}>'.format(email.to) )
        self.server.connect( self._host, self._port )
        # self.server.starttls()
        self.server.ehlo()
        self.server.login( self._account, self._password)
        self.server.sendmail( self._account, email.to, msg.as_string())
        self.server.quit()

    def __format_addr( self, s ):
        name, addr = parseaddr(s)
        return formataddr( ( Header(name, 'utf-8').encode(), addr ) )

class EmailManager(object):
    """Used for generating ``Email`` database record and sending email
    """
    def __init__(self, config, database, *args, **kwargs):
        # self.config = config
        email = config.email
        self.db = database
        self._test = config.debug
        self._host     = email[ 'host' ]
        if self._host == 'smtp host':
            logger.warning( '[Email] Please Place Your smtp account into config file')
            import sys
            sys.exit( 1 )
        if( self._host == None ) :
            raise ValueError

        port = email.get( 'port' ) or 465
        self.sender = MailSender(self._host, email[ 'account' ], email[ 'password' ], port)
        self._process_pending = False
        self._sendQueue = Queue()

    def add_email(self, to, subject, content, type = None ):
        """append an email object to queue, the mail will be sent at proper time,
        Empty subject or content is not allowed.
        """
        # it may be better if the email pattern is checked
        if len(to) == 0:
            logger.warning("Remote address not set in an email")
            return
        if len(subject) == 0:
            logger.warning("Empty subject found in an email")
            return
        if len(content) == 0:
            logger.warning("Empty content found in an email")
            return

        email = Email(
            to=to,
            subject=subject,
            content=content,
            type=type,
            remark="pending")
        # db.session.add(email)
        self._sendQueue.put(email)
    
    def send(self, email: Email):
        self._sendQueue.put(email)

    def checkRemain(self):
        """Check it if there exists any pending emails that need to be sent
        """
        if self._sendQueue.empty():
            self._checkPending()
            return
        email = self._sendQueue.get()
        self.__send(email, commit=True)

    def _checkPending(self):
        if self._process_pending == True:
            return
        self._process_pending = True
        emails = Email.query.filter_by(remark='pending').all()
        for e in emails:
            self.__send(e)
        self.db.session.commit()
        self._process_pending = False

    def __send(self, email: Email, commit = False):
        """Do the action of sending email, 
        ``commit`` set True to invoke db session commit.
        """
        self.db.session.add(email)
        logger.debug("Geting email and ready to send it")
        if self._test:
            email.remark = "debugNotSend"
            logger.info("Email generated but not send: {}".format(email))
        else:
            self.sender.reset_server()
            self.sender.send(email)
            email.remark = "sent"
            logger.debug("Email generated and send: {}".format(email))

        if commit:
            self.db.session.commit()


def registerApi(api):
    from flask import jsonify 
    @api.route_admin('/email/getAll', methods=['GET', 'POST'])
    def admin_getAllEmails():
        emails = Email.query.all()
        em = [e.to_dict() for e in emails]
        return jsonify({'status': 'success', 'emails': em})

    @api.route_admin('/email/getPage', methods=['GET', 'POST'])
    def admin_getPageEmails():
        page, per_page = api.getPageArgs()
        emails = Email.query.paginate(page=page, per_page=per_page)
        return jsonify({'status': 'success', 'emails': emails})
    logger.info("Api registered")



# -*- coding: utf-8 -*-

from flask import Blueprint, render_template

vuefront = Blueprint('vue', __name__, 
    template_folder="../vuetemplates", 
    static_folder="../vuetemplates/vstatic",
    static_url_path="/vstatic" )

@vuefront.route('/home', methods = [ 'GET' ], defaults= { 'path': ''} )
@vuefront.route('/home/<path:path>', methods = [ 'GET' ])
def view_home(path):
    return render_template( 'home.html' )

@vuefront.route('/user', methods = [ 'GET' ], defaults= { 'path': ''} )
@vuefront.route('/user/<path:path>', methods = [ 'GET' ])
def view_user(path):
    return render_template( 'user.html' )

@vuefront.route('/admin', methods = [ 'GET' ], defaults= { 'path': ''} )
@vuefront.route('/admin/<path:path>', methods = [ 'GET' ])
def view_admin(path):
    return render_template( 'admin.html' )

from flask import redirect

def init(app):
    @app.route("/", methods = [ 'GET' ] )
    def app_home_index():
        return redirect( '/home' )
    app.register_blueprint( vuefront )
    app.logger.info("[Plugin VueFront] is Ready")



# -*- coding: utf-8 -*-
"""Provide database model and routes as core plugin
"""

from .web_settings import WebguiSetting, registerApi as registerApiSetting
from .web_user import User, UserType, registerApi as registerApiUser,\
    getSuperManager, isManager, getCurrentUser
from .mail import EmailManager, registerApi as registerApiMail
from .routes import api

from sspymgr.manager import Manager
app = None

def init(iapp: Manager):
    global app
    app = iapp
    registerApiSetting(api, app)
    registerApiUser(api, app)
    app.m_emailManager = EmailManager(app.m_config, app.m_db)
    registerApiMail(api)


# -*- coding: utf-8 -*-
"""Description: replace sessions location, all neccessary session information will
be stored in database on your server. So it may be safer to store information within
session. It will clear all expired session record in database every time the SSPYMGR
starts (not interval check so the number of records may get bigger and bigger, it 
will be fixed some other time)

Author: BriFuture

Modified: 2019/03/10 20:07
"""

from .models import DB
from flask.sessions import SessionMixin, SessionInterface
from werkzeug import CallbackDict
from uuid import uuid4
from datetime import timedelta, datetime

class Sessions( DB.Model ):
    __tablename__ = 'sessions'
    sid = DB.Column( DB.String( 255 ), primary_key = True )
    expired = DB.Column( DB.DateTime )
    session = DB.Column( DB.Text )

    def __init__( self, **kwargs ):
        super().__init__( **kwargs )
        if 'sid' in kwargs:
            self.sid = kwargs[ 'sid' ]
        
        if 'expired' in kwargs:
            self.expired = kwargs[ 'expired' ]
        else:
            self.expired = datetime.now() + timedelta( minutes = 10 ) # from now

        if 'session' in kwargs:
            self.session = kwargs[ 'session' ]

    def __repr__( self ):
        return '<Session %s %s %s>' % ( self.sid, self.session, self.expired )

class DbSession( CallbackDict, SessionMixin ):
    def __init__( self, initial = None, sid = None, new = False ):
        def on_update( self ):
            self.modified = True

        CallbackDict.__init__( self, initial, on_update )
        self.sid = sid
        self.new = new
        self.modified = False

# def total_seconds( days, seconds = 0 ):
#     return days * 60 * 60 * 24 + seconds
import json

class DatabaseSessionInterface( SessionInterface ):
    serializer = json
    session_class = DbSession

    def __init__( self, database, prefix = 'session:' ):
        self.db = database
        self.prefix = prefix

    def generate_sid( self ):
        return str( uuid4() )
    
    def get_expiration_time( self, app, session ):
        if session.permanent:
            return datetime.now() + app.permanent_session_lifetime
        
        # keep session for 10 minutes
        return datetime.now() + timedelta( minutes = 10 )

    def open_session( self, app, request ):
        sid = request.cookies.get( app.session_cookie_name ) # 尝试从请求中获取存储的 cookie（与 session 相关的 cookie）
        if not sid:
            # 没有从 cookie 中找到相关 sid，说明客户端是第一次访问，需要构造 sid
            sid = self.generate_sid()
            # create new Session representation
            dbrecord = Sessions( sid = sid )
            # db.session.add( dbrecord )
            sc = self.session_class( sid = sid, new = True )
            # print( 'open_new: ', dbrecord, sc.new )
            return sc
        
        # if client have a session_cookie, take relative one from database
        dbrecord = Sessions.query.filter_by( sid = sid ).first()
        
        # print( "open: ", dbrecord )

        if dbrecord is None:
            return self.session_class( sid = sid ) # in case of wrong sid in session_cookie
        else:
            return self.session_class( self.serializer.loads( dbrecord.session ), sid = sid )

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain( app )
        # print( "save ", session, dbrecord )
        # print( 'save new ', session.new )
        dbrecord = Sessions.query.filter_by( sid = session.sid ).first()
        if not session:  
            # session content is empty
            if session.modified:
                # execute deletion if session is empty but modified
                self.db.session.delete( dbrecord )
                self.db.session.commit()
                response.delete_cookie( app.session_cookie_name, domain = domain )

            return 
        
        # update or insert
        if dbrecord is None:
            # in case database is clean
            dbrecord = Sessions( sid = session.sid )
            self.db.session.add( dbrecord )
            
        dbrecord.session = self.serializer.dumps( session )
        sexp = self.get_expiration_time( app, session )
        dbrecord.expired = sexp
        # print( "save:", dbrecord )
        self.db.session.commit()
        response.set_cookie( app.session_cookie_name, session.sid,
            expires = sexp, httponly = True, domain = domain )

import hashlib
from .globalfuncs import getRandomCode

def clearExpired(db):
    now = datetime.now()
    Sessions.query.filter(now > Sessions.expired).delete()
    db.session.commit()
   
def replaceSessionInterface(app):
    app.session_interface = DatabaseSessionInterface( database = app.m_db )
    # app.secret_key = hashlib.sha1( getRandomCode( 24 ).encode() ).hexdigest()
    app.m_events.on('afterCreateDb', clearExpired)
    app.secret_key = hashlib.sha1( 'test'.encode() ).hexdigest()


# -*- coding: utf-8 -*-
"""Description: This module is used for integrating shadowsocks in sspymgr, it defines the configuration file
that shadowsocks server should read when it starts. It also provides a very convenient way to start shadowsocks
server in a seperated thread without effecting the webserver of sspymgr. If only the shadowsocks server should 
be launched, invoke the ``main`` functin with wait keyword to be True in case the main thread finishes and then
the single thread with shadowsocks server running within it ends up. Please note: the output of the shadowsocks
server is redirected to a LOG file whose name is 'shadowsocks.log', placed under Universal Log Dir For sspymgr.

Author: BriFuture

Date: 2019/03/19
"""
import sys, os

from sspymgr import tr
from sspymgr.path_helper import createLoggerFile, DATA_DIR
from sspymgr.sscontroller import ssAddr

_config_file = '/server-multi-passwd.json'
def ssConfigFile(filename = '/server-multi-passwd.json'):
    """Get the configuration file for shadowsocks server, 
    TODO  able to change the config file when command-line options specified
    """
    config = DATA_DIR + filename
    if( not os.path.exists( config ) ):
        
        DEFAULT_CONFIG = """{
    "server": "0.0.0.0",
    "server_port": 0,
    "timeout": 300,
    "method": "aes-256-cfb",
    "fast_open": false
}
"""
        with open( config, 'w', encoding="utf-8") as f:
            f.write( DEFAULT_CONFIG )
    return config


def _runSS():
    """Deprecated, the launch process of shadowsocks server needs some configuration on the arguments option 
    which would modify ``sys.argv``, so it may contain potential problem if the sspymgr is started with some
    options.

    At the very begging of this project when there is no logger and command-line argument options for sspymgr,
    this method works fine, but as the project gets completed, this simple way to start shadowsocks server is
    no longer satisfied with the demands. 
    """
    mgr_addr = ssAddr(True)
    args = '--manager-address %s -c %s' % ( mgr_addr, ssConfigFile() )
    sys.argv[1:] = args.split( ' ' )   
    from shadowsocks import server
    server.main()

def new_runSS():
    """An effective way to start shadowsocks server with all the outputs of it redirected into a log file
    named `shadowsocks.log` under `~/.sspy-mgr/logs`.
    """
    import logging
    fileHandler = createLoggerFile('shadowsocks')
    logger = logging.getLogger()
    logger.addHandler(fileHandler)
    logger.setLevel(logging.INFO)

    from shadowsocks import shell, daemon, eventloop, manager
    logger.info( 'Start running shadowsock server')
    shell.check_python()
    logging.basicConfig(level=logging.INFO,
        format='%(asctime)s %(levelname)-8s %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')

    config_path = ssConfigFile()
    logger.info(tr('Loading configuration from {}').format(config_path))
    with open(config_path, 'rb') as f:
        try:
            config = shell.parse_json_in_str(f.read().decode('utf8'))
        except ValueError as e:
            logger.error("found an error in config.json: {}".format(e.message))
            sys.exit(1)
    config['manager_address'] = ssAddr(True)
    config['port_password'] = {}
    config['server_port'] = 0
    config['verbose'] = 2
    config['method'] = 'aes-256-cfb'

    daemon.daemon_exec(config)

    config['port_password'] = {}
    # server_port = config['server_port']
    # if type(server_port) == list:
    #     for a_server_port in server_port:
    #         config['port_password'][a_server_port] = config['password']
    # else:
    #     config['port_password'][str(server_port)] = config['password']

    logger.info('entering manager mode')
    manager.run(config)

def main(wait=False):
    """Set wait to true, the main process will wait shadowsocks process finished,
    it might be useful when webserver does not start.
    """
    ## The reason why Popen is not used is that the python interpreter should be specified to start the 
    ## sockserver. On some machines which contains both python2.x and python3.x, and python is linked to
    ## python2, that may cause some conflicts, so it's better to run the shadowsocks server in the sub
    ## Process and if sspymgr is launched normally, the sub process should be launched too if no error occurs.
    # import sys
    # from subprocess import Popen
    # with open( 'logfile.txt', 'ab' ) as file:
        # p = Popen( ['python', '-m', 'sockserver' ], stdin=sys.stdin, stdout=sys.stdout, stderr=file  ) 
    
    from multiprocessing import Process
    p = Process( target = new_runSS )
    p.daemon = True
    p.start()
    if wait:
        p.join()
    from time import sleep
    sleep( 1 )

if __name__ == '__main__':
    new_runSS()



# -*- coding: utf-8 -*-
"""Description: Core Subsystem in sspymgr, it provides plugin feature for the application.
It's a module which relies on importlib, the builtin module in python, but still has some
space to make improvements. It contains Events, Tasks and PluginLoader.
Note: And there is a problem which has not be solved if your python version is 3.6, 
it will work fine. But if your python version is higher than 3.6, for example, 3.7, 
it may not load plugins as expected, for that case, you should manually annotate the 
code under ``load_plugins`` function.

TODO: simplify the plugin loader

Author: BriFuture

Modified: 2019/03/10 20:14
"""


from .path_helper import createLogger
logger = createLogger('pluginLoader', stream=False, logger_prefix="[PluginLoader]")

from gettext import gettext as tr

class Events(dict):
    def __getitem__(self,name):
        if name not in self:
            super().__setitem__(name, [])
        return super().__getitem__(name)

    def trigger(self, eventName: str, eventArgs = None, **kwargs):
        """trigger event `eventName`
        """
        try:
            for e in self[eventName]:
                e(eventArgs, **kwargs)
        except Exception as exc:
            logger.warn("Exception at 'beforeRegisterRoutes: {}".format(exc))
        
    def on(self, eventName: str, callback ):
        """append callback function for certain event
        """
        self[eventName].append(callback)

import schedule

class BackTask(object):
    def __init__(self, events):
        self.events = events

    def start(self):
        """this method makes sure that endless while loop running in a subthread and will not interfere webgui
        """
        from threading import Thread
        
        mt = Thread( target= self.__loop_monitor )
        mt.daemon = True
        mt.start()
        
    def __loop_monitor(self):
        import time
        self.events.trigger("task_schedule", schedule)
        logger.info( 'start background tasks....' )
        count = 0
        while True:
            # schedule.run_pending()
            try:
                schedule.run_pending()
            except Exception as e:
                logger.warn("Schedule running: {}".format(e))
            count += 1
            if count == 3600:
                logger.debug("One Hour Pass.")
                count = 0
            time.sleep( 1 )


def load_plugins(app):
    from pathlib import Path
    import os
    pluginPath = Path(os.path.dirname(__file__)) / "plugins"
    
    import sys
    sys.path.insert(0, str(pluginPath))

    # 过滤掉非插件文件
    plugins = []
    for file in pluginPath.iterdir():
        if file.name.startswith('__'):
            continue
        if file.is_dir():
            dir_content = [x.name for x in file.iterdir()]
            if "__init__.py" in dir_content:
                plugins.append(file)
            dir_content.clear()
        elif file.is_file() and file.suffix == ".py":
            plugins.append(file)
    
    import importlib

    # filter those plugins whose dependencies are not meet
    logger.info(tr("Starting filtering plugins..."))
    plugin_module = []
    no_dependences = []
    plugin_file_names = [plugin_file.stem for plugin_file in plugins]
    for plugin_file in plugins:
        try:
            name = plugin_file.stem
            # print(' ==>', name)
            if plugin_file.is_dir():
                # following works for python3.6
                pm = importlib.import_module("{}".format(name))
                # following works for python3.7
                # pm = importlib.import_module("{}/__init__".format(name))
            else:
                pm = importlib.import_module(name)
            pm.name = name

            if(not hasattr(pm, 'requirement') or len(pm.requirement) == 0):
                logger.debug(tr("Importing dependless plugin: {}").format(pm))
                no_dependences.append(pm)
            else:
                meet = True
                for r in pm.requirement:
                    if r not in plugin_file_names:
                        meet = False
                if meet:
                    plugin_module.append(pm)
                    logger.debug(tr("Importing depend plugin: {}").format(pm))
        except Exception as exc:
            logger.warn("Load plugin `{}` Error: {}".format(plugin_file, exc))
    plugins.clear()
    importlib.invalidate_caches()
    logger.info(tr("Plugins is filtered, count of plugins that dont rely on others: {}, count of plugins that rely on others: {}")\
        .format(len(no_dependences), len(plugin_module)))

    num_all_plugins = len(no_dependences) + len(plugin_module)
    logger.info(tr("Preparing to initialize plugins..."))
    app.m_plugins = []
    num_loaded = 0
    level_loaded = 0
    while(num_loaded < num_all_plugins and level_loaded < 3):
        num_loaded += __do_load_plugin(app, plugin_module, no_dependences)
        level_loaded += 1
    logger.info(tr("Initialization completed, {} plugins has been loaded").format(num_loaded))

def __do_load_plugin(app, plugin_module, no_dependences):
    # 首先加载没有依赖的插件
    # logger.debug("no_dependces: {}".format(no_dependences))
    num_loaded = 0
    for nd in no_dependences:
        try:
            logger.debug(tr("Init plugin: {}").format(nd))
            nd.init(app)
            # logger.debug("load: {}".format(nd))
            app.m_plugins.append(nd)
            num_loaded += 1
        except Exception as exc:
            logger.warn("Error when init plugin `{}` : {}".format(nd, exc))

    # logger.debug("""loaded: {}, \nno_depend: {}, 
    #     plugin_module: {}""".format(app.m_plugins, no_dependences, plugin_module))
    no_dependences.clear()

    # 满足依赖关系的插件
    for depend in plugin_module:
        for pm in app.m_plugins:
        # for req in pm.requirement:
            if pm.name in depend.requirement:
                depend.requirement.remove(pm.name)

    no_dependences.clear()
    tmp_plugins = []
    for pm in plugin_module:
        if len(pm.requirement) > 0:
            tmp_plugins.append(pm)
        else:
            no_dependences.append(pm)
    plugin_module.clear()
    for pm in tmp_plugins:
        plugin_module.append(pm)
    # logger.debug("no_depend: {}".format(no_dependences))
    return num_loaded
            



# -*- coding: utf-8 -*-

from sspymgr import createLogger
logger = createLogger(
    "plugin_sscontroller", stream=False, logger_prefix="[Plugin sscontroller]")
from sspymgr import getRandomCode
from datetime import datetime, timedelta

app = None

from .flow_stats import AccountsChecker
from .flow_stats import FlowStats

import threading
def run_threaded(job_func):
    job_thread = threading.Thread(target=job_func)
    job_thread.start()

def scheduleTasks(schedule):
    recorder = FlowStats(app.m_db)
    app.m_sscontroller.setStats(recorder.recordFlow)
    # schedule.every(10).seconds.do(run_threaded, recorder.record5minToDb)
    schedule.every().minute.at(":57").do(run_threaded, recorder.record5minToDb)
    schedule.every().day.at("23:59").do(run_threaded,  recorder.record1dayToDb)
    accChecker = AccountsChecker(app)
    schedule.every().minutes.do(accChecker.checkAllAccounts)
    logger.debug("Tasks scheduled done")


from sspymgr.core import WebguiSetting
from sspymgr.sscontroller import availiable_port

def afterSingup(params: dict):
    port = availiable_port(app.m_db)
    password = getRandomCode(8)
    days = WebguiSetting.getSetting(key='web_signup_trial_days',\
            default_value=1, default_type="Number").getTypedValue()
    expire = timedelta(days=days) + datetime.now()
    added = app.m_sscontroller.add_account(
        port,
        password,
        userId=params['userId'],
        flow=params['flow'],
        expire=expire)
    logger.debug("signup accounts added: {}, params: {}".format(added, params))


from .ssc_routes import registerRoutes


def registerApi(api):
    registerRoutes(api, app)


from sspymgr import Manager


def init(iapp: Manager):
    global app
    app = iapp
    app.m_events.on('beforeRegisterApi', registerApi)
    app.m_events.on('task_schedule', scheduleTasks)
    app.m_events.on('webuser_signup_success', afterSingup)
    logger.info("inited")